'''
Unit tests for request module
'''

import ooipy.request.hydrophone_request as hyd_request
import ooipy.request.ctd_request as ctd_request
from ooipy.hydrophone.basic import HydrophoneData
from ooipy.ctd.basic import CtdData
import datetime
import numpy as np


def test_get_acoustic_data():
    # 1. case: 100% data coverage
    start_time = datetime.datetime(2017,3,10,0,0,0)
    end_time = datetime.datetime(2017,3,10,0,5,0)
    node = 'PC01A'

    data = hyd_request.get_acoustic_data(start_time, end_time, node)

    assert type(data) == HydrophoneData
    assert type(data.data) == np.ndarray

    diff_start = abs((start_time - data.stats.starttime.datetime).microseconds)
    diff_end = abs((end_time - data.stats.endtime.datetime).microseconds)

    assert diff_start <= 100
    assert diff_end <= 100

    # 2. case: 0% data coverage
    start_time = datetime.datetime(2017,10,10,15,30,0)
    end_time = datetime.datetime(2017,10,10,15,35,0)
    node = 'LJ01C'

    data = hyd_request.get_acoustic_data(start_time, end_time, node, append=False)

    assert data is None

    # 3. case: partial data coverage (data available until 15:17:50)
    start_time = datetime.datetime(2017,10,10,15,15,0)
    end_time = datetime.datetime(2017,10,10,15,20,0)
    node = 'LJ01C'

    data = hyd_request.get_acoustic_data(start_time, end_time, node, append=False)

    assert type(data) == HydrophoneData
    assert type(data.data) == np.ndarray

    diff_start = abs((start_time - data.stats.starttime.datetime).microseconds)
    diff_end = abs((end_time - data.stats.endtime.datetime).microseconds)
    assert diff_start <= 100
    assert diff_end > 100

    # 4. case: 0% data coverage for entire day (directory does not exists)
    start_time = datetime.datetime(2019,11,1,0,0,0)
    end_time = datetime.datetime(2019,11,1,0,5,0)
    node = 'LJ01D'

    data = hyd_request.get_acoustic_data(start_time, end_time, node, append=False)

    assert data is None

def test_hydrophone_node_names():
    node_arr = ['Oregon_Shelf_Base_Seafloor', 'Oregon_Slope_Base_Seafloor',
                'Oregon_Slope_Base_Shallow', 'Axial_Base_Shallow',
                'Oregon_Offshore_Base_Seafloor', 'Axial_Base_Seafloor']
    node_id_arr = ['LJ01D', 'LJ01A', 'PC01A', 'PC03A', 'LJ01C', 'LJ03A']

    starttime = datetime.datetime(2017,3,20,0,0,0) # time of first sample
    endtime = datetime.datetime(2017,3,20,0,0,1) # time of last sample

    for item in node_arr:
        hyd_data = hyd_request.get_acoustic_data(starttime, endtime, node=item)
        assert hyd_data.stats.location in node_id_arr
            
    node_arr = ['Slope_Base', 'Southern_Hydrate', 'Axial_Base',
                'Central_Caldera', 'Eastern_Caldera']
    node_id_arr = ['HYSB1', 'HYS14', 'AXBA1', 'AXCC1', 'AXEC2']

    for item in node_arr:
        hyd_data = hyd_request.get_acoustic_data_LF(starttime, endtime, node=item)
        assert hyd_data.stats.location in node_id_arr

# TODO: need to figure out how to do the OOI authentification on the GitHub VM
#def test_get_ctd_data():
#    start = datetime.datetime(2016, 12, 17, 2, 0, 0)
#    end = datetime.datetime(2016, 12, 17, 3, 0, 0)
#    ctd_data = ctd_request.get_ctd_data(start, end, 'oregon_offshore', limit=120)
#    
#    assert type(ctd_data) == HydrophoneData
#    assert len(ctd_data.raw_data) > 0


'''
This modules provides functions for plotting spectrograms and power
spectral density estimates. It extends the matplotlib.pyplot.plot
function.
'''

# Import all dependancies
import datetime
import numpy as np
import matplotlib
from matplotlib import pyplot as plt
from ooipy.hydrophone.basic import Spectrogram, Psd, HydrophoneData
from ooipy.ctd.basic import CtdProfile
from matplotlib.colors import Normalize
import matplotlib.dates as mdates
from obspy.core import UTCDateTime


def plot(*args, scalex=True, scaley=True, data=None, **kwargs):
    """
    An extension to the matplotlib.pyplot.plot function that allows for
    the nice plotting of :class:`ooipy.hydrophone.basic.Spectrogram`
    and :class:`ooipy.hydrophone.basic.Psd` objects. For a description
    of the input parameters, please refer to the matplotlib
    documentation.

    >>> from ooipy.request import hydrophone_request
    >>> from ooipy.tools import ooiplotlib as ooiplt

    >>> # pull OOI data and compute spectrogram and PSD
    >>> start_time = datetime.datetime(2017,3,10,0,0,0)
    >>> end_time = datetime.datetime(2017,3,10,0,5,0)
    >>> node = 'PC01A'
    >>> hydrophone_data = hydrophone_request.get_acoustic_data(
            start_time, end_time, node)
    >>> hydrophone_data.compute_spectrogram()
    >>> hydrophone_data.compute_psd_welch()

    >>> # plot spectrogram and change some default plot settings
    >>> ooiplt.plot(hydrophone_data.spectrogram)
    >>> plt.title('my spectrogram')
    >>> plt.ylim([0, 20000])
    >>> plt.show()

    >>> # plot PSD and chnage some default plot settings
    >>> ooiplt.plot(hydrophone_data.psd)
    >>> plt.title('my PSD')
    >>> plt.show()

    Parameters
    ----------
    args :
        object (either :class:`ooipy.hydrophone.basic.Spectrogram` or
        :class:`ooipy.hydrophone.basic.Psd`) or array to be plotted
    scalex :
        see matplotlib documentation
    scaley :
        see matplotlib doccumentation
    data :
        see matplotlib doccumentation
    kwargs :
        see matplotlib doccumentation,
        :func:`ooipy.tools.ooiplotlib.plot_spectrogram`, and
        :func:`ooipy.tools.ooiplotlib.plot_psd` for possible arguments
    """
    for arg in args:
        if isinstance(arg, Spectrogram):
            plot_spectrogram(arg, **kwargs)
        elif isinstance(arg, Psd):
            plot_psd(arg, **kwargs)
        elif isinstance(arg, HydrophoneData):
            plot_timeseries(arg, **kwargs)
        elif isinstance(arg, CtdProfile):
            plot_ctd_profile(arg, **kwargs)
        else:
            plt.gca().plot(arg, scalex=scalex, scaley=scaley,
                           **({"data": data} if data is not None else {}),
                           **kwargs)


def plot_spectrogram(spec_obj, **kwargs):
    """
    Plot a :class:`ooipy.hydrophone.basic.Spectrogram` object using the
    matplotlib package.

    Parameters
    ----------
    spec_obj : :class:`ooipy.hydrophone.basic.Spectrogram`
        spectrogram object to be plotted
    kwargs :
        See matplotlib doccumentation for list of arguments. Additional
        arguments are

        * plot : bool
            If False, figure will be closed. Can save time if only
            saving but not plotting is desired. Default is True
        * save : bool
            If True, figure will be saved under **filename**. Default is
            False
        * filename : str
            filename of figure if saved. Default is "spectrogram.png"
        * xlabel_rot : int or float
            rotation angle (deg) of x-labels. Default is 70
        * xlabel_format : str
            format of the xlabel if the time array contains datetime
            objects
        * fmin : int or float
            minimum frequency. Default is 0
        * fmax : int or float
            maximum frequency. Default is 32000
        * vmin : int or float
            lower limit of level axis (colormap). Default is 20
        * vmax : int or float
            upper limit of level axis (colormap). Default is 80
        * vdelta : int or float
            resolution of level axis (colormap). Default is 1
        * vdelta_cbar : int or float
            label distance of colorbar. Default is 5
        * figsize : (int, int)
            width and height of figure, Default is (16, 9)
        * res_reduction_time : int
            reduction factor of time domain resolution. This can
            facilitate faster plotting of large spectroagm objects.
            Default is 1 (no reduction)
        * res_reduction_freq : int
            reduction factor of frequency domain resolution. This can
            facilitate faster plotting of large spectroagm objects.
            Default is 1 (no reduction)
    """
    # check for keys
    if 'plot' not in kwargs:
        kwargs['plot'] = True
    if 'save' not in kwargs:
        kwargs['save'] = False
    if 'filename' not in kwargs:
        kwargs['filename'] = 'spectrogram.png'
    if 'title' not in kwargs:
        kwargs['title'] = 'Spectrogram'
    if 'xlabel' not in kwargs:
        kwargs['xlabel'] = 'time'
    if 'xlabel_rot' not in kwargs:
        kwargs['xlabel_rot'] = 70
    if 'xlabel_format' not in kwargs:
        kwargs['xlabel_format'] = '%y-%m-%d %H:%M'
    if 'ylabel' not in kwargs:
        kwargs['ylabel'] = 'frequency'
    if 'fmin' not in kwargs:
        kwargs['fmin'] = 0.0
    if 'fmax' not in kwargs:
        kwargs['fmax'] = 32000.0
    if 'vmin' not in kwargs:
        kwargs['vmin'] = 20.0
    if 'vmax' not in kwargs:
        kwargs['vmax'] = 80.0
    if 'vdelta' not in kwargs:
        kwargs['vdelta'] = 1.0
    if 'vdelta_cbar' not in kwargs:
        kwargs['vdelta_cbar'] = 5.0
    if 'figsize' not in kwargs:
        kwargs['figsize'] = (16, 9)
    if 'res_reduction_time' not in kwargs:
        kwargs['res_reduction_time'] = 1
    if 'res_reduction_freq' not in kwargs:
        kwargs['res_reduction_freq'] = 1

    # set backend for plotting/saving:
    if not kwargs['plot']:
        matplotlib.use('Agg')
    font = {'size': 22}
    matplotlib.rc('font', **font)

    # reduce resolution in time and frequency
    v = spec_obj.values[::kwargs['res_reduction_time'],
                        ::kwargs['res_reduction_freq']]
    if len(spec_obj.time) != len(spec_obj.values):
        t = np.linspace(0, len(spec_obj.values) - 1, int(len(spec_obj.values)
                        / kwargs['res_reduction_time']))
    else:
        t = spec_obj.time[::kwargs['res_reduction_time']]
    if len(spec_obj.freq) != len(spec_obj.values[0]):
        f = np.linspace(0, len(spec_obj.values[0]) - 1,
                        int(len(spec_obj.values[0]) /
                        kwargs['res_reduction_freq']))
    else:
        f = spec_obj.freq[::kwargs['res_reduction_freq']]

    # plot spectrogram object
    cbarticks = np.arange(kwargs['vmin'], kwargs['vmax'] + kwargs['vdelta'],
                          kwargs['vdelta'])
    fig, ax = plt.subplots(figsize=kwargs['figsize'])
    im = ax.contourf(t, f, np.transpose(v), cbarticks,
                     norm=Normalize(vmin=kwargs['vmin'], vmax=kwargs['vmax']),
                     cmap=plt.cm.jet, **kwargs)
    plt.ylabel(kwargs['ylabel'])
    plt.xlabel(kwargs['xlabel'])
    plt.ylim([kwargs['fmin'], kwargs['fmax']])
    plt.xticks(rotation=kwargs['xlabel_rot'])
    plt.title(kwargs['title'])
    plt.colorbar(im, ax=ax, ticks=np.arange(kwargs['vmin'], kwargs['vmax'] +
                                            kwargs['vdelta'],
                                            kwargs['vdelta_cbar']))
    plt.tick_params(axis='y')

    if isinstance(t[0], datetime.datetime) or isinstance(t[0], UTCDateTime):
        ax.xaxis.set_major_formatter(mdates.DateFormatter(
            kwargs['xlabel_format']))

    if kwargs['save']:
        plt.savefig(kwargs['filename'], bbox_inches='tight')

    if not kwargs['plot']:
        plt.close(fig)


def plot_psd(psd_obj, **kwargs):
    """
    Plot a :class:`ooipy.hydrophone.basic.Psd` object using the
    matplotlib package.

    Parameters
    ----------
    spec_obj : :class:`ooipy.hydrophone.basic.Psd`
        Psd object to be plotted
    kwargs :
        See matplotlib doccumentation for list of arguments. Additional
        arguments are

        * plot : bool
            If False, figure will be closed. Can save time if only
            saving but not plotting is desired. Default is True
        * save : bool
            If True, figure will be saved under **filename**. Default is
            False
        * new_fig : bool
            If True, matplotlib will create a new fugure. Default is
            True
        * filename : str
            filename of figure if saved. Default is "spectrogram.png"
        * xlabel_rot : int or float
            rotation angle (deg) of x-labels. Default is 70
        * fmin : int or float
            minimum frequency. Default is 0
        * fmax : int or float
            maximum frequency. Default is 32000
        * vmin : int or float
            lower limit of level axis (colormap). Default is 20
        * vmax : int or float
            upper limit of level axis (colormap). Default is 80
        * figsize : (int, int)
            width and height of figure. Default is (16, 9)
    """

    # check for keys
    if 'plot' not in kwargs:
        kwargs['plot'] = True
    if 'save' not in kwargs:
        kwargs['save'] = False
    if 'new_fig' not in kwargs:
        kwargs['new_fig'] = True
    if 'filename' not in kwargs:
        kwargs['filename'] = 'psd.png'
    if 'title' not in kwargs:
        kwargs['title'] = 'PSD'
    if 'xlabel' not in kwargs:
        kwargs['xlabel'] = 'frequency'
    if 'xlabel_rot' not in kwargs:
        kwargs['xlabel_rot'] = 0
    if 'ylabel' not in kwargs:
        kwargs['ylabel'] = 'spectral level'
    if 'fmin' not in kwargs:
        kwargs['fmin'] = 0.0
    if 'fmax' not in kwargs:
        kwargs['fmax'] = 32000.0
    if 'vmin' not in kwargs:
        kwargs['vmin'] = 20.0
    if 'vmax' not in kwargs:
        kwargs['vmax'] = 80.0
    if 'figsize' not in kwargs:
        kwargs['figsize'] = (16, 9)

    # set backend for plotting/saving:
    if not kwargs['plot']:
        matplotlib.use('Agg')
    font = {'size': 22}
    matplotlib.rc('font', **font)

    if len(psd_obj.freq) != len(psd_obj.values):
        f = np.linspace(0, len(psd_obj.values) - 1, len(psd_obj.values))
    else:
        f = psd_obj.freq

    # plot PSD object
    if kwargs['new_fig']:
        fig, ax = plt.subplots(figsize=kwargs['figsize'])
    plt.semilogx(f, psd_obj.values)
    plt.ylabel(kwargs['ylabel'])
    plt.xlabel(kwargs['xlabel'])
    plt.xlim([kwargs['fmin'], kwargs['fmax']])
    plt.ylim([kwargs['vmin'], kwargs['vmax']])
    plt.xticks(rotation=kwargs['xlabel_rot'])
    plt.title(kwargs['title'])
    plt.grid(True)

    if kwargs['save']:
        plt.savefig(kwargs['filename'], bbox_inches='tight')

    if not kwargs['plot']:
        plt.close(fig)


def plot_timeseries(HydData_obj, **kwargs):
    """
    Plot a :class:`ooipy.hydrophone.basic.HydrophoneData` object using the
    matplotlib package.

    Parameters
    ----------
    HydData_obj : :class:`ooipy.hydrophone.basic.HydrophoneData`
        Hydrophone Data to be plotted
    kwargs :
        See matplotlib doccumentation for list of arguments. Additional
        arguments are

        * plot : bool
            If False, figure will be closed. Can save time if only
            saving but not plotting is desired. Default is True
        * save : bool
            If True, figure will be saved under **filename**. Default is
            False
        * filename : str
            filename of figure if saved. Default is "spectrogram.png"
        * xlabel_rot : int or float
            rotation angle (deg) of x-labels. Default is 70
        * xlabel_format : str
            format of the xlabel if the time array contains datetime
            objects
        * figsize : (int, int)
            width and height of figure, Default is (16, 9)
        * res_reduction_time : int
            reduction factor of time domain resolution. This can
            facilitate faster plotting of large spectroagm objects.
            Default is 1 (no reduction)
    """
    # check for keys
    if 'plot' not in kwargs:
        kwargs['plot'] = True
    if 'save' not in kwargs:
        kwargs['save'] = False
    if 'filename' not in kwargs:
        kwargs['filename'] = 'timeseries.png'
    if 'title' not in kwargs:
        kwargs['title'] = 'Time Series Plot'
    if 'xlabel' not in kwargs:
        kwargs['xlabel'] = None
    if 'xlabel_rot' not in kwargs:
        kwargs['xlabel_rot'] = 0
    if 'xlabel_format' not in kwargs:
        kwargs['xlabel_format'] = '%y-%m-%d %H:%M'
    if 'ylabel' not in kwargs:
        kwargs['ylabel'] = 'Amplitude'
    if 'figsize' not in kwargs:
        kwargs['figsize'] = (16, 9)
    if 'res_reduction_time' not in kwargs:
        kwargs['res_reduction_time'] = 1
    if 'dpi' not in kwargs:
        kwargs['dpi'] = 100
    if 'fontsize' not in kwargs:
        kwargs['fontsize'] = 22

    # create array of datetime.datetime instances
    import pandas as pd

    start = pd.Timestamp(HydData_obj.stats.starttime.datetime)
    end = pd.Timestamp(HydData_obj.stats.endtime.datetime)
    tpd = np.linspace(start.value, end.value, HydData_obj.stats.npts)
    tpd = pd.to_datetime(tpd)

    time = tpd.to_numpy()

    # set backend for plotting/saving:
    if not kwargs['plot']:
        matplotlib.use('Agg')
    font = {'size': kwargs['fontsize']}
    matplotlib.rc('font', **font)

    # reduce resolution in time
    data_reduce = HydData_obj.data[::kwargs['res_reduction_time']]
    if len(time) != len(HydData_obj.data):
        t = np.linspace(0, len(HydData_obj.data) - 1, int(len(HydData_obj.data)
                        / kwargs['res_reduction_time']))
    else:
        t = time[::kwargs['res_reduction_time']]

    # plot HydrophoneData object
    fig, ax = plt.subplots(figsize=kwargs['figsize'])
    ax.plot(t, data_reduce)
    plt.ylabel(kwargs['ylabel'])
    plt.xlabel(kwargs['xlabel'])
    plt.xticks(rotation=kwargs['xlabel_rot'])
    plt.title(kwargs['title'])

    plt.tick_params(axis='y')

    if isinstance(t[0], datetime.datetime) or isinstance(t[0], UTCDateTime):
        ax.xaxis.set_major_formatter(mdates.DateFormatter(
            kwargs['xlabel_format']))

    if kwargs['save']:
        plt.savefig(kwargs['filename'], bbox_inches='tight', dpi=kwargs['dpi'])
    if not kwargs['plot']:
        plt.close(fig)


def plot_ctd_profile(ctd_profile, **kwargs):
    """
    Plot a :class:`ooipy.ctd.basic.CtdProfile` object using the
    matplotlib package.

    Parameters
    ----------
    ctd_profile : :class:`ooipy.ctd.basic.CtdProfile`
        CtdProfile object to be plotted
    kwargs :
        See matplotlib doccumentation for list of arguments. Additional
        arguments are

        * plot : bool
            If False, figure will be closed. Can save time if only
            saving but not plotting is desired. Default is True
        * save : bool
            If True, figure will be saved under **filename**. Default is
            False
        * new_fig : bool
            If True, matplotlib will create a new fugure. Default is
            True
        * filename : str
            filename of figure if saved. Default is "spectrogram.png"
        * figsize : (int, int)
            width and height of figure. Default is (16, 9)
        * title : str
            Title of plot. Default is 'CTD profile'
        * xlabel : str
            x-axis label of plot. Default is 'parameter'
        * ylabel : str
            y-axis label of plot. Default is 'depth'
        * show_variance : bool
            Indicates whether the variance should be plotted or not.
            Default is True
        * min_depth : int or float
            upper limit of vertical axis (depth axis). Default is the
            maximum of max(min(ctd_profile.depth_mean) - 10, 0)
        * max_depth : int or float
            lower limit of vertical axis (depth axis). Default is
            max(ctd_profile.depth_mean) + 10
    """

    # check for keys
    if 'plot' not in kwargs:
        kwargs['plot'] = True
    if 'save' not in kwargs:
        kwargs['save'] = False
    if 'new_fig' not in kwargs:
        kwargs['new_fig'] = True
    if 'filename' not in kwargs:
        kwargs['filename'] = 'ctd_profile.png'
    if 'title' not in kwargs:
        kwargs['title'] = 'CTD profile'
    if 'xlabel' not in kwargs:
        kwargs['xlabel'] = 'parameter'
    if 'ylabel' not in kwargs:
        kwargs['ylabel'] = 'depth'
    if 'figsize' not in kwargs:
        kwargs['figsize'] = (16, 9)
    if 'show_variance' not in kwargs:
        kwargs['show_variance'] = True
    if 'linestyle' not in kwargs:
        kwargs['linestyle'] = 'dashed'
    if 'marker' not in kwargs:
        kwargs['marker'] = 'o'
    if 'markersize' not in kwargs:
        kwargs['markersize'] = 5
    if 'color' not in kwargs:
        kwargs['color'] = 'black'
    if 'alpha' not in kwargs:
        kwargs['alpha'] = 0.5
    if 'var_color' not in kwargs:
        kwargs['var_color'] = 'gray'
    if 'min_depth' not in kwargs:
        kwargs['min_depth'] = max(ctd_profile.depth_mean[0] - 10, 0)
    if 'max_depth' not in kwargs:
        kwargs['max_depth'] = np.nanmax(ctd_profile.depth_mean) + 10

    # set backend for plotting/saving:
    if not kwargs['plot']:
        matplotlib.use('Agg')
    font = {'size': 22}
    matplotlib.rc('font', **font)

    # plot PSD object
    if kwargs['new_fig']:
        fig, ax = plt.subplots(figsize=kwargs['figsize'])
    plt.plot(ctd_profile.parameter_mean, ctd_profile.depth_mean,
             linestyle=kwargs['linestyle'], marker=kwargs['marker'],
             markersize=kwargs['markersize'], color=kwargs['color'])
    if kwargs['show_variance']:
        y1 = ctd_profile.parameter_mean - 2 * \
            np.sqrt(ctd_profile.parameter_var)
        y2 = ctd_profile.parameter_mean + 2 * \
            np.sqrt(ctd_profile.parameter_var)
        plt.fill_betweenx(ctd_profile.depth_mean, y1, y2,
                          alpha=kwargs['alpha'], color=kwargs['var_color'])
    plt.ylim([kwargs['max_depth'], kwargs['min_depth']])
    plt.ylabel(kwargs['ylabel'])
    plt.xlabel(kwargs['xlabel'])
    plt.title(kwargs['title'])
    plt.grid(True)

    if kwargs['save']:
        plt.savefig(kwargs['filename'], bbox_inches='tight')

    if not kwargs['plot']:
        plt.close(fig)



"""
Module for hydrophone (acoustic) data objects

The HydrophoneData objects inherits from obspy.Trace. Furthermore,
methods for computing spectrograms and power spectral densities are
added.
"""
import numpy as np
import json
from matplotlib import pyplot as plt
from obspy import Trace
from obspy.core import UTCDateTime
from scipy import signal
from scipy.interpolate import interp1d
import matplotlib.dates as mdates
import matplotlib.colors as colors
import matplotlib
import datetime
import multiprocessing as mp
import pickle
from scipy.io import wavfile
import warnings
import ooipy
import pandas as pd
import os


class HydrophoneData(Trace):
    """
    Object that stores hydrophone data

    Attributes
    ----------
    spectrogram : Spectrogram
        spectrogram of HydrophoneData.data.data.
        Spectral level, time, and frequency bins can be accessed by
        spectrogram.values, spectrogram.time, and spectrogram.freq
    psd : Psd
        power spectral density estimate of HydrophoneData.data.data.
        Spectral level and frequency bins can be accessed by
        psd.values and psd.freq
    psd_list : list of :class:`.Psd`
        the data object is divided into N segments and for each
        segment a separate power spectral density estimate is computed and
        stored in psd_list. psd_list is computed by compute_psd_welch_mp
    type : str
        Either 'broadband' or 'low_frequency' specifies the type of hydrophone
        that the date is from.

    """

    def __init__(self, data=np.array([]), header=None, node=''):

        super().__init__(data, header)
        self.stats.location = node_id(node)

        self.spectrogram = None
        self.psd = None
        self.psd_list = None
        self.type = None

    # TODO: use correct frequency response for all hydrophones
    def frequency_calibration(self, N):
        # TODO
        """
        Apply a frequency dependent sensitivity correction to the
        acoustic data based on the information from the calibration
        sheets.
        TODO Add for all broadband hydrophones
        !!! Currently only implemented for Oregon Offshore Base Seafloor
        and Oregon Shelf Base Seafloor hydrophone. For all other
        hydrophones, an average sensitivity of -169dBV/1uPa is assumed
        !!!

        Parameters
        ----------
        N : int
            length of the data segment

        Returns
        -------
        output_array : np.array
            array with correction coefficient for every frequency
        """
        # Load calibation file and get appropriate calibration info
        filename = os.path.dirname(ooipy.__file__) + \
            '/hydrophone/calibration_by_assetID.csv'
        # Use deployment CSV to determine asset_ID
        assetID = self.get_asset_ID()

        # load calibration data as pandas dataframe
        cal_by_assetID = pd.read_csv(filename, header=[0, 1])

        f_calib = cal_by_assetID[assetID]['Freq (kHz)'].to_numpy() * 1000
        sens_calib_0 = cal_by_assetID[assetID]['0 phase'].to_numpy()
        sens_calib_90 = cal_by_assetID[assetID]['90 phase'].to_numpy()
        sens_calib = 0.5 * (sens_calib_0 + sens_calib_90)
        f = np.linspace(0, round(self.stats.sampling_rate / 2), N)

        # Convert calibration to correct units
        if round(self.stats.sampling_rate) == 200:
            sens_calib = 20 * np.log10(sens_calib * 1e-6)
        elif round(self.stats.sampling_rate) == 64000:
            sens_calib = sens_calib + 128.9
        else:
            raise Exception('Invalid sampling rate')

        sens_interpolated = interp1d(f_calib, sens_calib)

        f_calib = sens_interpolated(f)
        return f_calib

    def compute_spectrogram(self, win='hann', L=4096, avg_time=None,
                            overlap=0.5, verbose=True):
        """
        Compute spectrogram of acoustic signal. For each time step of the
        spectrogram either a modified periodogram (avg_time=None)
        or a power spectral density estimate using Welch's method with median
        averaging is computed.

        Parameters
        ----------
        win : str, optional
            Window function used to taper the data. See scipy.signal.get_window
            for a list of possible window functions (Default is Hann-window.)
        L : int, optional
            Length of each data block for computing the FFT (Default is 4096).
        avg_time : float, optional
            Time in seconds that is covered in one time step of the
            spectrogram. Default value is None and one time step covers L
            samples. If the signal covers a long time period it is recommended
            to use a higher value for avg_time to avoid memory overflows and
            to facilitate visualization.
        overlap : float, optional
            Percentage of overlap between adjacent blocks if Welch's method is
            used. Parameter is ignored if avg_time is None. (Default is 50%)
        verbose : bool, optional
            If true (defult), exception messages and some comments are printed.

        Returns
        -------
        Spectrogram
            A Spectrogram object that contains time and frequency bins as
            well as corresponding values. If no noise date is available,
            None is returned.
        """
        specgram = []
        time = []

        if any(self.data) is None:
            if verbose:
                print('Data object is empty. Spectrogram cannot be computed')
            self.spectrogram = None
            return None

        # sampling frequency
        fs = self.stats.sampling_rate

        # number of time steps
        if avg_time is None:
            nbins = int((len(self.data) - L) / ((1 - overlap) * L)) + 1
        else:
            nbins = int(np.ceil(len(self.data) / (avg_time * fs)))

        # compute spectrogram. For avg_time=None
        # (periodogram for each time step), the last data samples are ignored
        # if len(noise[0].data) != k * L
        if avg_time is None:
            n_hop = int(L * (1 - overlap))
            for n in range(nbins):
                f, Pxx = signal.periodogram(
                    x=self.data[n * n_hop:n * n_hop + L],
                    fs=fs, window=win)
                if len(Pxx) != int(L / 2) + 1:
                    if verbose:
                        print('Error while computing periodogram for segment',
                              n)
                    self.spectrogram = None
                    return None
                else:
                    sense_corr = -self.frequency_calibration(
                        int(L / 2 + 1))

                    Pxx = 10 * np.log10(Pxx * np.power(10, sense_corr / 10))

                    specgram.append(Pxx)
                    time.append(self.stats.starttime.datetime
                                + datetime.timedelta(seconds=n * L / fs))

        else:
            for n in range(nbins - 1):
                f, Pxx = signal.welch(
                    x=self.data[n * int(fs * avg_time):
                                (n + 1) * int(fs * avg_time)],
                    fs=fs, window=win, nperseg=L, noverlap=int(L * overlap),
                    nfft=L, average='median')

                if len(Pxx) != int(L / 2) + 1:
                    if verbose:
                        print('Error while computing '
                              'Welch estimate for segment', n)
                    self.spectrogram = None
                    return None
                else:
                    sense_corr = -self.frequency_calibration(
                        int(L / 2 + 1))

                    Pxx = 10 * np.log10(Pxx * np.power(10, sense_corr / 10))
                    specgram.append(Pxx)
                    time.append(self.stats.starttime.datetime
                                + datetime.timedelta(seconds=n * avg_time))

            # compute PSD for residual segment
            # if segment has more than L samples
            if len(self.data[int((nbins - 1) * fs * avg_time):]) >= L:
                f, Pxx = signal.welch(
                    x=self.data[int((nbins - 1) * fs * avg_time):],
                    fs=fs, window=win, nperseg=L, noverlap=int(L * overlap),
                    nfft=L, average='median')
                if len(Pxx) != int(L / 2) + 1:
                    if verbose:
                        print(
                            'Error while computing Welch '
                            'estimate residual segment')
                    self.spectrogram = None
                    return None
                else:
                    sense_corr = -self.frequency_calibration(
                        int(L / 2 + 1))

                    Pxx = 10 * np.log10(Pxx * np.power(10, sense_corr / 10))
                    specgram.append(Pxx)
                    time.append(self.stats.starttime.datetime
                                + datetime.timedelta(seconds=(nbins - 1)
                                                     * avg_time))

        if len(time) == 0:
            if verbose:
                print('Spectrogram does not contain any data')
            self.spectrogram = None
            return None
        else:
            self.spectrogram = Spectrogram(np.array(time), np.array(f),
                                           np.array(specgram))
            return self.spectrogram

    def compute_spectrogram_mp(self, n_process=None, win='hann', L=4096,
                               avg_time=None, overlap=0.5, verbose=True):
        """
        Same as function compute_spectrogram but using multiprocessing.
        This function is intended to be used when analyzing large data sets.

        Parameters
        ----------
        n_process : int, optional
            Number of processes in the pool. None (default) means that
            n_process is equal to the number of CPU cores.
        win : str, optional
            Window function used to taper the data.
            See scipy.signal.get_window for a list of possible window functions
            (Default is Hann-window.)
        L : int, optional
            Length of each data block for computing the FFT (Default is 4096).
        avg_time : float, optional
            Time in seconds that is covered in one time step of the
            spectrogram. Default value is None and one time step covers L
            samples. If the signal covers a long time period it is recommended
            to use a higher value for avg_time to avoid memory overflows and
            to facilitate visualization.
        overlap : float, optional
            Percentage of overlap between adjacent blocks if Welch's method
            is used. Parameter is ignored if avg_time is None. (Default is 50%)
        verbose : bool, optional
            If true (defult), exception messages and some comments are printed.

        Returns
        -------
        Spectrogram
            A Spectrogram object that contains time and frequency bins as well
            as corresponding values. If no noise date is available,
            None is returned.
        """

        # create array with N start and end time values
        if n_process is None:
            N = mp.cpu_count()
        else:
            N = n_process

        ooi_hyd_data_list = []
        seconds_per_process = (self.stats.endtime - self.stats.starttime) / N
        for k in range(N - 1):
            starttime = self.stats.starttime + datetime.timedelta(
                seconds=k * seconds_per_process)
            endtime = self.stats.starttime + datetime.timedelta(
                seconds=(k + 1) * seconds_per_process)
            temp_slice = self.slice(starttime=UTCDateTime(starttime),
                                    endtime=UTCDateTime(endtime))
            tmp_obj = HydrophoneData(data=temp_slice.data,
                                     header=temp_slice.stats,
                                     node=self.stats.location)
            ooi_hyd_data_list.append((tmp_obj, win, L, avg_time, overlap))

        starttime = self.stats.starttime + datetime.timedelta(
            seconds=(N - 1) * seconds_per_process)
        temp_slice = self.slice(starttime=UTCDateTime(starttime),
                                endtime=UTCDateTime(self.stats.endtime))
        tmp_obj = HydrophoneData(data=temp_slice.data, header=temp_slice.stats,
                                 node=self.stats.location)
        ooi_hyd_data_list.append((tmp_obj, win, L, avg_time, overlap))

        with mp.get_context("spawn").Pool(n_process) as p:
            try:
                specgram_list = p.starmap(_spectrogram_mp_helper,
                                          ooi_hyd_data_list)
                # concatenate all small spectrograms to
                # obtain final spectrogram
                specgram = []
                time_specgram = []
                for i in range(len(specgram_list)):
                    time_specgram.extend(specgram_list[i].time)
                    specgram.extend(specgram_list[i].values)
                self.spectrogram = Spectrogram(np.array(time_specgram),
                                               specgram_list[0].freq,
                                               np.array(specgram))
                return self.spectrogram
            except Exception:
                if verbose:
                    print('Cannot compute spectrogram')
                self.spectrogram = None
                return None

    def compute_psd_welch(self, win='hann', L=4096, overlap=0.5,
                          avg_method='median', interpolate=None, scale='log',
                          verbose=True):
        """
        Compute power spectral density estimates of noise data using
        Welch's method.

        Parameters
        ----------
        win : str, optional
            Window function used to taper the data. See scipy.signal.get_window
            for a list of possible window functions (Default is Hann-window.)
        L : int, optional
            Length of each data block for computing the FFT (Default is 4096).
        overlap : float, optional
            Percentage of overlap between adjacent blocks if Welch's method is
            used. Parameter is ignored if avg_time is None. (Default is 50%)
        avg_method : str, optional
            Method for averaging the periodograms when using Welch's method.
            Either 'mean' or 'median' (default) can be used
        interpolate : float, optional
            Resolution in frequency domain in Hz. If None (default), the
            resolution will be sampling frequency fs divided by L. If
            interpolate is smaller than fs/L, the PSD will be interpolated
            using zero-padding
        scale : str, optional
            If 'log' (default) PSD in logarithmic scale (dB re 1µPa^2/H) is
            returned. If 'lin', PSD in linear scale
            (1µPa^2/H) is returned
        verbose : bool, optional
            If true (default), exception messages and some comments are
            printed.

        Returns
        -------
        Psd
            A Psd object that contains frequency bins and PSD values. If no
            noise date is available, None is returned.
        """
        # get noise data segment for each entry in rain_event
        # each noise data segment contains usually 1 min of data
        if any(self.data) is None:
            if verbose:
                print('Data object is empty. PSD cannot be computed')
            self.psd = None
            return None
        fs = self.stats.sampling_rate

        # compute nfft if zero padding is desired
        if interpolate is not None:
            if fs / L > interpolate:
                nfft = int(fs / interpolate)
            else:
                nfft = L
        else:
            nfft = L

        # compute Welch median for entire data segment
        f, Pxx = signal.welch(x=self.data, fs=fs, window=win, nperseg=L,
                              noverlap=int(L * overlap),
                              nfft=nfft, average=avg_method)

        if len(Pxx) != int(nfft / 2) + 1:
            if verbose:
                print('PSD cannot be computed.')
            self.psd = None
            return None

        sense_corr = -self.frequency_calibration(
            int(nfft / 2 + 1))
        print(sense_corr)
        if scale == 'log':
            Pxx = 10 * np.log10(Pxx * np.power(10, sense_corr / 10))
        elif scale == 'lin':
            Pxx = Pxx * np.power(10, sense_corr / 10)
        else:
            raise Exception('scale has to be either "lin" or "log".')

        self.psd = Psd(f, Pxx)
        return self.psd

    def compute_psd_welch_mp(self, split, n_process=None, win='hann', L=4096,
                             overlap=0.5, avg_method='median',
                             interpolate=None, scale='log', verbose=True):
        """
        Same as compute_psd_welch but using the multiprocessing library.

        Parameters
        ----------
        split : int, float, or array of datetime.datetime
            Time period for each PSD estimate. The time between start_time and
            end_time is split into parts of length
            split seconds (if float). The last segment can be shorter than
            split seconds. Alternatively split can
            be set as an list, where each entry is a start-end time tuple.
        n_process : int, optional
            Number of processes in the pool. None (default) means that
            n_process is equal to the number
            of CPU cores.
        win : str, optional
            Window function used to taper the data. See scipy.signal.get_window
            for a list of possible window functions (Default is Hann-window.)
        L : int, optional
            Length of each data block for computing the FFT (Default is 4096).
        overlap : float, optional
            Percentage of overlap between adjacent blocks if Welch's method is
            used. Parameter is ignored if avg_time is None. (Default is 50%)
        avg_method : str, optional
            Method for averaging the periodograms when using Welch's method.
            Either 'mean' or 'median' (default) can be used
        interpolate : float, optional
            Resolution in frequency domain in Hz. If None (default), the
            resolution will be sampling frequency fs divided by L. If
            interpolate is smaller than fs/L, the PSD will be interpolated
            using zero-padding
        scale : str, optional
            If 'log' (default) PSD in logarithmic scale (dB re 1µPa^2/H) is
            returned. If 'lin', PSD in linear scale (1µPa^2/H) is returned
        verbose : bool, optional
            If true (default), exception messages and some comments are
            printed.

        Returns
        -------
        list of Psd
            A list of Psd objects where each entry represents the PSD of the
            respective noise segment. If no noise data is available, None is
            returned.
        """

        # create array with N start and end time values
        if n_process is None:
            n_process = mp.cpu_count()

        ooi_hyd_data_list = []
        # do segmentation from scratch
        if isinstance(split, int) or isinstance(split, float):
            n_seg = int(
                np.ceil((self.stats.endtime - self.stats.starttime) / split))

            seconds_per_process = \
                (self.stats.endtime - self.stats.starttime) / n_seg

            for k in range(n_seg - 1):
                starttime = self.stats.starttime + datetime.timedelta(
                    seconds=k * seconds_per_process)
                endtime = self.stats.starttime + datetime.timedelta(
                    seconds=(k + 1) * seconds_per_process)
                temp_slice = self.slice(starttime=UTCDateTime(starttime),
                                        endtime=UTCDateTime(endtime))
                tmp_obj = HydrophoneData(data=temp_slice.data,
                                         header=temp_slice.stats,
                                         node=self.stats.location)
                ooi_hyd_data_list.append(
                    (tmp_obj, win, L, overlap, avg_method, interpolate, scale))

            # treat last segment separately as its length may differ from other
            # segments
            starttime = self.stats.starttime + datetime.timedelta(
                seconds=(n_seg - 1) * seconds_per_process)
            temp_slice = self.slice(starttime=UTCDateTime(starttime),
                                    endtime=UTCDateTime(self.stats.endtime))
            tmp_obj = HydrophoneData(data=temp_slice.data,
                                     header=temp_slice.stats,
                                     node=self.stats.location)
            ooi_hyd_data_list.append(
                (tmp_obj, win, L, overlap, avg_method, interpolate, scale))
        # use segmentation specified by split
        else:
            ooi_hyd_data_list = []
            for row in split:
                temp_slice = self.slice(starttime=UTCDateTime(row[0]),
                                        endtime=UTCDateTime(row[1]))
                tmp_obj = HydrophoneData(data=temp_slice.data,
                                         header=temp_slice.stats,
                                         node=self.stats.location)
                ooi_hyd_data_list.append(
                    (tmp_obj, win, L, overlap, avg_method, interpolate, scale))

        with mp.get_context("spawn").Pool(n_process) as p:
            try:
                self.psd_list = p.starmap(_psd_mp_helper, ooi_hyd_data_list)
            except Exception:
                if verbose:
                    print('Cannot compute PSd list')
                self.psd_list = None

        return self.psd_list

    def wav_write(self, filename, norm=False, new_sample_rate=None):
        '''
        method that stores HydrophoneData into .wav file

        Parameters
        ----------
        filename : str
            filename to store .wav file as
        norm : bool
            specifices whether data should be normalized to 1
        new_sample_rate : float
            specifices new sample rate of wav file to be saved. (Resampling is
            done with scipy.signal.resample()). Default is None which keeps
            original sample rate of data.
        '''
        if norm:
            data = self.data / np.abs(np.max(self.data))
        else:
            data = self.data

        if new_sample_rate is None:
            sampling_rate = self.stats.sampling_rate
        else:
            if new_sample_rate > self.stats.sampling_rate:
                upsamp_fac = new_sample_rate / self.stats.sampling_rate
                new_npts = self.stats.npts * upsamp_fac
                data = signal.resample(data, int(new_npts))
                sampling_rate = new_sample_rate
            elif new_sample_rate == self.stats.sampling_rate:
                warnings.warn('New sample rate is same as original data. '
                              'No resampling done.')
                sampling_rate = self.stats.sampling_rate
            elif new_sample_rate < self.stats.sampling_rate:
                warnings.warn('New sample rate is lower than original sample'
                              ' rate. Chebychev 1 anti-aliasing filter used')
                if self.stats.sampling_rate % new_sample_rate != 0:
                    raise Exception(
                        'New Sample Rate is not factor of original sample rate'
                    )
                else:
                    data = signal.decimate(
                        data, int(self.stats.sampling_rate / new_sample_rate))
                    sampling_rate = new_sample_rate

        wavfile.write(filename, int(sampling_rate), data)

    def get_asset_ID(self):
        '''
        get_asset_ID returns the hydrophone asset ID for a given data sample.
        This data can be foun `here <https://raw.githubusercontent.com/
        OOI-CabledArray/deployments/main/HYDBBA_deployments.csv'>`_ for
        broadband hydrophones. Since Low frequency hydrophones remain
        constant with location and time, if the hydrophone is low frequency,
        the node ID is returned
        '''
        # Low frequency hydrophone
        if round(self.stats.sampling_rate) == 200:
            asset_ID = self.stats.location

        elif round(self.stats.sampling_rate) == 64000:
            url = 'https://raw.githubusercontent.com/OOI-CabledArray/' \
                'deployments/main/HYDBBA_deployments.csv'
            hyd_df = pd.read_csv(url)

            # LJ01D'Oregon Shelf Base Seafloor
            if self.stats.location == 'LJ01D':
                ref = 'CE02SHBP-LJ01D-11-HYDBBA106'
            # LJ01AOregon Slope Base Seafloor
            if self.stats.location == 'LJ01A':
                ref = 'RS01SLBS-LJ01A-09-HYDBBA102'
            # Oregan Slope Base Shallow
            if self.stats.location == 'PC01A':
                ref = 'RS01SBPS-PC01A-08-HYDBBA103'
            # Axial Base Shallow Profiler
            if self.stats.location == 'PC03A':
                ref = 'RS03AXPS-PC03A-08-HYDBBA303'
            # Oregon Offshore Base Seafloor
            if self.stats.location == 'LJ01C':
                ref = 'CE04OSBP-LJ01C-11-HYDBBA105'
            # Axial Base Seafloor
            if self.stats.location == 'LJ03A':
                ref = 'RS03AXBS-LJ03A-09-HYDBBA302'

            hyd_df['referenceDesignator']

            df_ref = hyd_df.loc[hyd_df['referenceDesignator'] == ref]

            df_start = df_ref.loc[(df_ref['startTime'] < self.stats.starttime)
                                  & (df_ref['endTime']
                                  > self.stats.starttime)]
            df_end = df_ref.loc[(df_ref['startTime'] < self.stats.endtime) &
                                (df_ref['endTime'] > self.stats.endtime)]

            if df_start.index.to_numpy() == df_end.index.to_numpy():
                idx = df_start.index.to_numpy()
                asset_ID = df_start['assetID'][int(idx)]
            else:
                raise Exception('Hydrophone Data involves multiple'
                                'deployments. Feature to be added later')
        else:
            raise Exception('Invalid hydrophone sampling rate')

        return asset_ID


def node_id(node):
    '''
    mapping of name of hydrophone node to ID

    Parameter
    ---------
    node : str
        name or ID of the hydrophone node

    Returns
    -------
    str
        ID of hydrophone node
    '''
    # broadband hydrophones
    if node == 'Oregon_Shelf_Base_Seafloor' or node == 'LJ01D':
        return 'LJ01D'
    if node == 'Oregon_Slope_Base_Seafloor' or node == 'LJ01A':
        return 'LJ01A'
    if node == 'Oregon_Slope_Base_Shallow' or node == 'PC01A':
        return 'PC01A'
    if node == 'Axial_Base_Shallow' or node == 'PC03A':
        return 'PC03A'
    if node == 'Oregon_Offshore_Base_Seafloor' or node == 'LJ01C':
        return 'LJ01C'
    if node == 'Axial_Base_Seafloor' or node == 'LJ03A':
        return 'LJ03A'

    # low frequency hydrophones
    if node == 'Slope_Base' or node == 'HYSB1':
        return 'HYSB1'
    if node == 'Southern_Hydrate' or node == 'HYS14':
        return 'HYS14'
    if node == 'Axial_Base' or node == 'AXBA1':
        return 'AXBA1'
    if node == 'Central_Caldera' or node == 'AXCC1':
        return 'AXCC1'
    if node == 'Eastern_Caldera' or node == 'AXEC2':
        return 'AXEC2'

    else:
        print('No node exists for name or ID ' + node)
        return ''


def node_name(node):
    '''
    mapping of ID of hydrophone node to name

    Parameter
    ---------
    node : str
        ID or name of the hydrophone node

    Returns
    -------
    str
        name of hydrophone node
    '''
    # broadband hydrophones
    if node == 'Oregon_Shelf_Base_Seafloor' or node == 'LJ01D':
        return 'Oregon_Shelf_Base_Seafloor'
    if node == 'Oregon_Slope_Base_Seafloor' or node == 'LJ01A':
        return 'Oregon_Slope_Base_Seafloor'
    if node == 'Oregon_Slope_Base_Shallow' or node == 'PC01A':
        return 'Oregon_Slope_Base_Shallow'
    if node == 'Axial_Base_Shallow' or node == 'PC03A':
        return 'Axial_Base_Shallow'
    if node == 'Oregon_Offshore_Base_Seafloor' or node == 'LJ01C':
        return 'Oregon_Offshore_Base_Seafloor'
    if node == 'Axial_Base_Seafloor' or node == 'LJ03A':
        return 'Axial_Base_Seafloor'

    # low frequency hydrophones
    if node == 'Slope_Base' or node == 'HYSB1':
        return 'Slope_Base'
    if node == 'Southern_Hydrate' or node == 'HYS14':
        return 'Southern_Hydrate'
    if node == 'Axial_Base' or node == 'AXBA1':
        return 'Axial_Base'
    if node == 'Central_Caldera' or node == 'AXCC1':
        return 'Central_Caldera'
    if node == 'Eastern_Caldera' or node == 'AXEC2':
        return 'Eastern_Caldera'

    else:
        print('No node exists for ID or name ' + node)
        return ''


def _spectrogram_mp_helper(ooi_hyd_data_obj, win, L, avg_time, overlap):
    """
    Helper function for compute_spectrogram_mp
    """
    ooi_hyd_data_obj.compute_spectrogram(win, L, avg_time, overlap)
    return ooi_hyd_data_obj.spectrogram


def _psd_mp_helper(ooi_hyd_data_obj, win, L, overlap, avg_method, interpolate,
                   scale):
    """
    Helper function for compute_psd_welch_mp
    """
    ooi_hyd_data_obj.compute_psd_welch(win, L, overlap, avg_method,
                                       interpolate, scale)
    return ooi_hyd_data_obj.psd


class Spectrogram:
    """
    A class used to represent a spectrogram object.

    Attributes
    ----------
    time : 1-D array of float or datetime objects
        Indices of time-bins of spectrogram.
    freq : 1-D array of float
        Indices of frequency-bins of spectrogram.
    values : 2-D array of float
        Values of the spectrogram. For each time-frequency-bin pair there has
        to be one entry in values. That is, if time has  length N and freq
        length M, values is a NxM array.

    """

    def __init__(self, time, freq, values):
        self.time = time
        self.freq = freq
        self.values = values

    def visualize(self, plot_spec=True, save_spec=False,
                  filename='spectrogram.png', title='spectrogram',
                  xlabel='time', xlabel_rot=70, ylabel='frequency', fmin=0,
                  fmax=32000, vmin=20, vmax=80, vdelta=1.0,
                  vdelta_cbar=5, figsize=(16, 9), dpi=96, res_reduction_time=1,
                  res_reduction_freq=1, time_limits=None):
        """
        This function will be depreciated into a differnt module in the future.
        The current documentation might not be accurate.

        To plot spectrograms please see
        :meth:`ooipy.hydrophone.basic.plot_spectrogram`

        Basic visualization of spectrogram based on matplotlib. The function
        offers two options: Plot spectrogram in Python (plot_spec = True) and
        save spectrogram plot in directory (save_spec = True). Spectrograms are
        plotted in dB re 1µ Pa^2/Hz.

        plot_spec (bool): whether or not spectrogram is plotted using Python
        save_spec (bool): whether or not spectrogram plot is saved
        filename (str): directory where spectrogram plot is saved. Use ending
        ".png" or ".pdf" to save as PNG or PDF file. This value will be ignored
        if save_spec=False
        title (str): title of plot
        ylabel (str): label of vertical axis
        xlabel (str): label of horizontal axis
        xlabel_rot (float): rotation of xlabel. This is useful if xlabel are
        longer strings for example when using datetime.datetime objects.
        fmin (float): minimum frequency (unit same as f) that is displayed
        fmax (float): maximum frequency (unit same as f) that is displayed
        vmin (float): minimum value (dB) of spectrogram that is colored.
        All values below are displayed in white.
        vmax (float): maximum value (dB) of spectrogram that is colored.
        All values above are displayed in white.
        vdelta (float): color resolution
        vdelta_cbar (int): label ticks in colorbar are in vdelta_cbar steps
        figsize (tuple(int)): size of figure
        dpi (int): dots per inch
        time_limits : list
            specifices xlimits on spectrogram. List contains two
            datetime.datetime objects
        """
        import warnings
        raise warnings.warn('will be depricated in future. Please see '
                            'ooipy.tools.ooiplotlib.plot_spectrogram()')
        # set backend for plotting/saving:
        if not plot_spec:
            matplotlib.use('Agg')

        font = {'size': 22}
        matplotlib.rc('font', **font)

        v = self.values[::res_reduction_time, ::res_reduction_freq]

        if len(self.time) != len(self.values):
            t = np.linspace(0, len(self.values) - 1,
                            int(len(self.values) / res_reduction_time))
        else:
            t = self.time[::res_reduction_time]

        # Convert t to np.array of datetime.datetime
        if type(t[0]) == UTCDateTime:
            for k in range(len(t)):
                t[k] = t[k].datetime

        if len(self.freq) != len(self.values[0]):
            f = np.linspace(0, len(self.values[0]) - 1,
                            int(len(self.values[0]) / res_reduction_freq))
        else:
            f = self.freq[::res_reduction_freq]

        cbarticks = np.arange(vmin, vmax + vdelta, vdelta)
        fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
        im = ax.contourf(t, f, np.transpose(v), cbarticks,
                         norm=colors.Normalize(vmin=vmin, vmax=vmax),
                         cmap=plt.cm.jet)
        plt.ylabel(ylabel)
        plt.xlabel(xlabel)
        plt.ylim([fmin, fmax])
        if time_limits is not None:
            plt.xlim(time_limits)
        plt.xticks(rotation=xlabel_rot)
        plt.title(title)
        plt.colorbar(im, ax=ax,
                     ticks=np.arange(vmin, vmax + vdelta, vdelta_cbar))
        plt.tick_params(axis='y')

        if type(t[0]) == datetime.datetime:
            ax.xaxis.set_major_formatter(
                mdates.DateFormatter('%y-%m-%d %H:%M'))

        if save_spec:
            plt.savefig(filename, dpi=dpi, bbox_inches='tight')

        if plot_spec:
            plt.show()
        else:
            plt.close(fig)

    def save(self, filename='spectrogram.pickle'):
        """
        !!!!! This function will be moved into a different module in the
        future. The current documentation might not be accurate !!!!!

        Save spectrogram in pickle file.

        filename (str): directory where spectrogram data is saved. Ending has
        to be ".pickle".
        """

        dct = {
            't': self.time,
            'f': self.freq,
            'spectrogram': self.values
        }
        with open(filename, 'wb') as outfile:
            pickle.dump(dct, outfile)

    def plot(self, **kwargs):
        '''
        redirects to ooipy.ooiplotlib.plot_spectrogram()
        please see :meth:`ooipy.hydrophone.basic.plot_spectrogram`
        '''
        ooipy.tools.ooiplotlib.plot_spectrogram(self, **kwargs)


class Psd:
    """
    A calss used to represent a PSD object

    Attributes
    ----------
    freq : array of float
        Indices of frequency-bins of PSD.
    values : array of float
        Values of the PSD.

    TODO:
    Methods
    -------
    visualize(plot_psd=True, save_psd=False, filename='psd.png', title='PSD',
    xlabel='frequency', xlabel_rot=0, ylabel='spectral level', fmin=0,
    fmax=32, vmin=20, vmax=80, figsize=(16,9), dpi=96)
        Visualizes PSD estimate using matplotlib.
    save(filename='psd.json', ancillary_data=[], ancillary_data_label=[])
        Saves PSD estimate and ancillary data in .json file.
    """

    def __init__(self, freq, values):
        self.freq = freq
        self.values = values

    def visualize(self, plot_psd=True, save_psd=False, filename='psd.png',
                  title='PSD', xlabel='frequency',
                  xlabel_rot=0, ylabel='spectral level', fmin=0, fmax=32000,
                  vmin=20, vmax=80, figsize=(16, 9), dpi=96):
        """
        !!!!! This function will be moved into a different module in the
        future. The current documentation might not be accurate !!!!!

        Basic visualization of PSD estimate based on matplotlib. The function
        offers two options: Plot PSD in Python (plot_psd = True) and save PSD
        plot in directory (save_psd = True). PSDs are plotted in dB re 1µ
        Pa^2/Hz.

        plot_psd (bool): whether or not PSD is plotted using Python
        save_psd (bool): whether or not PSD plot is saved
        filename (str): directory where PSD plot is saved. Use ending ".png"
        or ".pdf" to save as PNG or PDF
            file. This value will be ignored if save_psd=False
        title (str): title of plot
        ylabel (str): label of vertical axis
        xlabel (str): label of horizontal axis
        xlabel_rot (float): rotation of xlabel. This is useful if xlabel are
        longer strings.
        fmin (float): minimum frequency (unit same as f) that is displayed
        fmax (float): maximum frequency (unit same as f) that is displayed
        vmin (float): minimum value (dB) of PSD.
        vmax (float): maximum value (dB) of PSD.
        figsize (tuple(int)): size of figure
        dpi (int): dots per inch
        """
        import warnings
        raise warnings.warn('will be depricated in future. Please see '
                            'ooipy.tools.ooiplotlib.plot_psd()')
        # set backend for plotting/saving:
        if not plot_psd:
            matplotlib.use('Agg')

        font = {'size': 22}
        matplotlib.rc('font', **font)

        if len(self.freq) != len(self.values):
            f = np.linspace(0, len(self.values) - 1, len(self.values))
        else:
            f = self.freq

        fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
        plt.semilogx(f, self.values)
        plt.ylabel(ylabel)
        plt.xlabel(xlabel)
        plt.xlim([fmin, fmax])
        plt.ylim([vmin, vmax])
        plt.xticks(rotation=xlabel_rot)
        plt.title(title)
        plt.grid(True)

        if save_psd:
            plt.savefig(filename, dpi=dpi, bbox_inches='tight')

        if plot_psd:
            plt.show()
        else:
            plt.close(fig)

    def save(self, filename='psd.json', ancillary_data=[],
             ancillary_data_label=[]):
        """
        !!!!! This function will be moved into a different module in the
        future. The current documentation might not be accurate !!!!!

        Save PSD estimates along with with ancillary data
        (stored in dictionary) in json file.

        filename (str): directory for saving the data
        ancillary_data ([array like]): list of ancillary data
        ancillary_data_label ([str]): labels for ancillary data used as keys
        in the output dictionary.
            Array has same length as ancillary_data array.
        """

        if len(self.freq) != len(self.values):
            f = np.linspace(0, len(self.values) - 1, len(self.values))
        else:
            f = self.freq

        if type(self.values) != list:
            values = self.values.tolist()

        if type(f) != list:
            f = f.tolist()

        dct = {
            'psd': values,
            'f': f
        }

        if len(ancillary_data) != 0:
            for i in range(len(ancillary_data)):
                if type(ancillary_data[i]) != list:
                    dct[ancillary_data_label[i]] = ancillary_data[i].tolist()
                else:
                    dct[ancillary_data_label[i]] = ancillary_data[i]

        with open(filename, 'w+') as outfile:
            json.dump(dct, outfile)

    def plot(self, **kwargs):
        '''
        redirects to ooipy.ooiplotlib.plot_psd()
        please see :meth:`ooipy.hydrophone.basic.plot_psd`
        '''
        ooipy.tools.ooiplotlib.plot_psd(self, **kwargs)



"""
Module for CTD data objects
"""

import numpy as np
import datetime
import ooipy


class CtdData():
    """
    Object that stores conductivity, temperature, depth (CTD) data, and
    provides functions for calculating sound speed, temperature,
    pressure, and salinity profiles. When a CtdData object is created
    and extract_parameters = True (default), then temperature, pressure,
    salinity, and time are automatically extracted from the raw data.

    Attributes
    ----------
    raw_data : list of dict
        list containing sample from CTD. Each sample is a dictionary
        containing all parameters measured by the CTD.
    temperature : numpy.ndarray
        array containing temperature samples in degree celsius.
    pressure : numpy.ndarray
        array containing pressure samples in dbar.
    salinity : numpy.ndarray
        array containing salinity samples in parts per thousand.
    depth : numpy.ndarray
        array containing depth samples in meter.
    density : numpy.ndarray
        array containing density samples in kg/cubic meter.
    conductivity : numpy.ndarray
        array containing conductivity samples in siemens/meter.
    sound_speed : numpy.ndarray
        array containing sound speed samples in meter/second.
    time : numpy.ndarray
        array containing time samples as datetime.datetime objects.
    sound_speed_profile : :class:`ooipy.ctd.basic.CtdProfile`
        object for sound speed profile.
    temperature_profile : :class:`ooipy.ctd.basic.CtdProfile`
        object for temperature profile.
    salinity_profile : :class:`ooipy.ctd.basic.CtdProfile`
        object for salinity profile.
    pressure_profile : :class:`ooipy.ctd.basic.CtdProfile`
        object for pressure profile.
    density_profile : :class:`ooipy.ctd.basic.CtdProfile`
        object for density profile.
    conductivity_profile : :class:`ooipy.ctd.basic.CtdProfile`
        object for conductivity profile.

    """

    def __init__(self, raw_data=None, extract_parameters=True):
        self.raw_data = raw_data

        if self.raw_data is not None and extract_parameters:
            self.temperature = \
                self.get_parameter_from_rawdata('temperature')
            self.pressure = self.get_parameter_from_rawdata('pressure')
            self.salinity = self.get_parameter_from_rawdata('salinity')
            self.time = self.get_parameter_from_rawdata('time')
        else:
            self.temperature = None
            self.pressure = None
            self.salinity = None
            self.time = None

        self.sound_speed = None
        self.depth = None
        self.density = None
        self.conductivity = None
        self.sound_speed_profile = None
        self.temperature_profile = None
        self.salinity_profile = None
        self.pressure_profile = None
        self.density_profile = None
        self.conductivity_profile = None

    def ntp_seconds_to_datetime(self, ntp_seconds):
        """
        Converts timestamp into dattime object.
        """
        ntp_epoch = datetime.datetime(1900, 1, 1)
        unix_epoch = datetime.datetime(1970, 1, 1)
        ntp_delta = (unix_epoch - ntp_epoch).total_seconds()
        return datetime.datetime.utcfromtimestamp(
            ntp_seconds - ntp_delta).replace(microsecond=0)

    def get_parameter_from_rawdata(self, parameter):
        """
        Extracts parameters from raw data dictionary.
        """
        param_arr = []
        for item in self.raw_data:
            if parameter == 'temperature':
                if 'seawater_temperature' in item:
                    param_arr.append(item['seawater_temperature'])
                elif 'temperature' in item:
                    param_arr.append(item['temperature'])
                else:
                    param_arr.append(item['temp'])
            if parameter == 'pressure':
                if 'ctdbp_no_seawater_pressure' in item:
                    param_arr.append(item['ctdbp_no_seawater_pressure'])
                elif 'seawater_pressure' in item:
                    param_arr.append(item['seawater_pressure'])
                else:
                    param_arr.append(item['pressure'])

            if parameter == 'salinity':
                if 'practical_salinity' in item:
                    param_arr.append(item['practical_salinity'])
                else:
                    param_arr.append(item['salinity'])

            if parameter == 'density':
                if 'seawater_density' in item:
                    param_arr.append(item['seawater_density'])
                else:
                    param_arr.append(item['density'])

            if parameter == 'conductivity':
                if 'ctdbp_no_seawater_conductivity' in item:
                    param_arr.append(item['ctdbp_no_seawater_conductivity'])
                elif 'seawater_conductivity' in item:
                    param_arr.append(item['seawater_conductivity'])
                else:
                    param_arr.append(item['conductivity'])

            if parameter == 'time':
                param_arr.append(self.ntp_seconds_to_datetime(
                    item['pk']['time']))

        return np.array(param_arr)

    def get_parameter(self, parameter):
        """
        Extension of get_parameters_from_rawdata. Also sound speed and
        depth can be requested.
        """
        if parameter in ['temperature', 'pressure', 'salinity', 'time',
                         'density', 'conductivity']:
            param = self.get_parameter_from_rawdata(parameter)
        elif parameter == 'sound_speed':
            param = self.calc_sound_speed()
        elif parameter == 'depth':
            param = self.calc_depth_from_pressure()
        else:
            param = None

        return param

    def calc_depth_from_pressure(self):
        """
        Calculates depth from pressure array
        """

        if self.pressure is None:
            self.pressure = self.get_parameter_from_rawdata('pressure')

        press_MPa = 0.01 * self.pressure

        # TODO: adapt for each hydrophone
        lat = 44.52757  # deg

        # Calculate gravity constant for given latitude
        g_phi = 9.780319 * (1 + 5.2788E-3 * (np.sin(np.deg2rad(lat))**2) +
                            2.36E-5 * (np.sin(np.deg2rad(lat))**4))

        # Calculate Depth for Pressure array
        self.depth = (9.72659e2 * press_MPa - 2.512e-1 * press_MPa**2 +
                      2.279e-4 * press_MPa ** 3 - 1.82e-7 * press_MPa**4) / \
                     (g_phi + 1.092e-4 * press_MPa)

        return self.depth

    def calc_sound_speed(self):
        """
        Calculates sound speed from temperature, salinity and pressure
        array. The equation for calculating the sound speed is from:
        Chen, C. T., & Millero, F. J. (1977). Speed of sound in seawater
        at high pressures. Journal of the Acoustical Society of America,
        62(5), 1129–1135. https://doi.org/10.1121/1.381646
        """
        if self.pressure is None:
            self.pressure = self.get_parameter_from_rawdata('pressure')
        if self.temperature is None:
            self.temperature = self.get_parameter_from_rawdata('temperature')
        if self.salinity is None:
            self.salinity = self.get_parameter_from_rawdata('salinity')

        press_MPa = 0.01 * self.pressure

        C00 = 1402.388
        A02 = 7.166E-5
        C01 = 5.03830
        A03 = 2.008E-6
        C02 = -5.81090E-2
        A04 = -3.21E-8
        C03 = 3.3432E-4
        A10 = 9.4742E-5
        C04 = -1.47797E-6
        A11 = -1.2583E-5
        C05 = 3.1419E-9
        A12 = -6.4928E-8
        C10 = 0.153563
        A13 = 1.0515E-8
        C11 = 6.8999E-4
        A14 = -2.0142E-10
        C12 = -8.1829E-6
        A20 = -3.9064E-7
        C13 = 1.3632E-7
        A21 = 9.1061E-9
        C14 = -6.1260E-10
        A22 = -1.6009E-10
        C20 = 3.1260E-5
        A23 = 7.994E-12

        C21 = -1.7111E-6
        A30 = 1.100E-10
        C22 = 2.5986E-8
        A31 = 6.651E-12
        C23 = -2.5353E-10
        A32 = -3.391E-13
        C24 = 1.0415E-12
        B00 = -1.922E-2
        C30 = -9.7729E-9
        B01 = -4.42E-5
        C31 = 3.8513E-10
        B10 = 7.3637E-5
        C32 = -2.3654E-12
        B11 = 1.7950E-7
        A00 = 1.389
        D00 = 1.727E-3
        A01 = -1.262E-2
        D10 = -7.9836E-6

        T = 3
        S = 1
        P = 700
        T = self.temperature
        S = self.salinity
        P = press_MPa * 10

        D = D00 + D10 * P
        B = B00 + B01 * T + (B10 + B11 * T) * P
        A = (A00 + A01 * T + A02 * T**2 + A03 * T**3 + A04 * T**4) + \
            (A10 + A11 * T + A12 * T**2 + A13 * T**3 + A14 * T**4) * P + \
            (A20 + A21 * T + A22 * T**2 + A23 * T**3) * P**2 + \
            (A30 + A31 * T + A32 * T**2) * P**3
        Cw = (C00 + C01 * T + C02 * T**2 +
              C03 * T**3 + C04 * T**4 + C05 * T**5) + \
            (C10 + C11 * T + C12 * T**2 + C13 * T**3 + C14 * T**4) * P + \
            (C20 + C21 * T + C22 * T**2 + C23 * T**3 + C24 * T**4) * P**2 + \
            (C30 + C31 * T + C32 * T**2) * P**3

        # Calculate Speed of Sound
        self.sound_speed = Cw + A * S + B * S**(3 / 2) + D * S**2

        return self.sound_speed

    def get_profile(self, max_depth, parameter):
        """
        Compute the profile for sound speed, temperature, pressure, or
        salinity over the vater column.

        Parameters
        ----------
        max_depth : int
            The profile will be computed from 0 meters to max_depth
            meters in 1-meter increments
        parameter : str
            * 'sound_speed'
            * 'temperature'
            * 'salinity'
            * 'pressure'
            * 'density'
            * 'conductivity'
        """

        param_dct = {}
        for k in range(max_depth):
            param_dct[str(k)] = {'param': [], 'd': []}

        param_arr = self.get_parameter(parameter)

        if self.depth is None:
            self.depth = self.get_parameter('depth')

        for d, p in zip(self.depth, param_arr):
            if str(int(d)) in param_dct:
                param_dct[str(int(d))]['d'].append(d)
                param_dct[str(int(d))]['param'].append(p)

        param_mean = []
        depth_mean = []
        param_var = []
        depth_var = []

        n_samp = []

        for key in param_dct:
            param_mean.append(np.mean(param_dct[key]['param']))
            depth_mean.append(np.mean(param_dct[key]['d']))
            param_var.append(np.var(param_dct[key]['param']))
            depth_var.append(np.var(param_dct[key]['d']))
            n_samp.append(len(param_dct[key]['d']))

        idx = np.argsort(depth_mean)

        depth_mean = np.array(depth_mean)[idx]
        param_mean = np.array(param_mean)[idx]
        depth_var = np.array(depth_var)[idx]
        param_var = np.array(param_var)[idx]
        n_samp = np.array(n_samp)[idx]

        param_profile = CtdProfile(param_mean, param_var, depth_mean,
                                   depth_var, n_samp)

        if parameter == 'temperature':
            self.temperature_profile = param_profile
        elif parameter == 'salinity':
            self.salinity_profile = param_profile
        elif parameter == 'pressure':
            self.pressure_profile = param_profile
        elif parameter == 'sound_speed':
            self.sound_speed_profile = param_profile
        elif parameter == 'density':
            self.density_profile = param_profile
        elif parameter == 'conductivity':
            self.conductivity_profile = param_profile

        return param_profile


class CtdProfile():
    """
    Simple object that stores a parameter profile over the water column.
    For each 1-meter interval, there is one data point in the profile.

    Attributes
    ----------
    parameter_mean : array of float
        mean of paramter within each 1-meter depth interval
    parameter_var : array of float
        variance of paramter within each 1-meter depth interval
    depth_mean : array of float
        mean of depth within each 1-meter depth interval
    depth_var : array of float
        variance of depth within each 1-meter depth interval
    n_samp : array of int
        number of samples within each 1-meter depth interval
    """

    def __init__(self, parameter_mean, parameter_var, depth_mean, depth_var,
                 n_samp):
        self.parameter_mean = parameter_mean
        self.parameter_var = parameter_var
        self.depth_mean = depth_mean
        self.depth_var = depth_var
        self.n_samp = n_samp

    def plot(self, **kwargs):
        '''
        redirects to ooipy.ooiplotlib.plot_ctd_profile()
        please see :meth:`ooipy.hydrophone.basic.plot_psd`
        '''
        ooipy.tools.ooiplotlib.plot_ctd_profile(self, **kwargs)

    def convert_to_ssp(self):
        '''
        converts to numpy array with correct format for arlpy simulation

        Returns
        -------
        ssp : numpy array
            2D numpy array containing sound speed profile column 0 is depth,
            column 1 is sound speed (in m/s)
        '''
        ssp = np.vstack((self.depth_mean, self.parameter_mean)).T
        # insert 0 depth term
        ssp = np.insert(ssp, 0, np.array((0, ssp[0, 1])), 0)

        # remove NaN Terms
        first_nan = np.where(np.isnan(ssp))[0][0]
        ssp = ssp[:first_nan - 1, :]

        return ssp



'''
This modules handles the downloading of OOI Data. As of current, the supported
OOI sensors include all broadband hydrophones (Fs = 64 kHz) and all low
frequency hydrophones (Fs = 200 Hz). All supported hydrophone nodes are listed
in the Hydrophone Nodes section below.

.. sidebar:: Hydrophone Request Jupyter Notebook

        For a demo of the hydrophone request module, see the `Hydrophone
        Request Jupyter Notebook <_static/test_request.html>`_.

Hydrophone Nodes
^^^^^^^^^^^^^^^^
* `Oregon Shelf Base Seafloor (Fs = 64 kHz)
<https://ooinet.oceanobservatories.org/data_access/
?search=CE02SHBP-LJ01D-11-HYDBBA106>`_
    * 'LJ01D'
* `Oregon Slope Base Seafloor (Fs = 64 kHz)
<https://ooinet.oceanobservatories.org/data_access/
?search=RS01SLBS-LJ01A-09-HYDBBA102>`_
    * 'LJ01A'
* `Slope Base Shallow (Fs = 64 kHz)
<https://ooinet.oceanobservatories.org/data_access/
?search=RS01SBPS-PC01A-08-HYDBBA103>`_
    * 'PC01A'
* `Axial Base Shallow Profiler (Fs = 64 kHz)
<https://ooinet.oceanobservatories.org/data_access/
?search=RS03AXPS-PC03A-08-HYDBBA303>`_
    * 'PC03A'
* `Offshore Base Seafloor (Fs = 64 kHz)
<https://ooinet.oceanobservatories.org/data_access/
?search=CE04OSBP-LJ01C-11-HYDBBA105>`_
    * 'LJ01C'
* `Axial Base Seafloor (Fs = 64 kHz)
<https://ooinet.oceanobservatories.org/data_access/
?search=RS03AXBS-LJ03A-09-HYDBBA302>`_
    * 'LJ03A'
* `Axial Base Seaflor (Fs = 200 Hz)
<https://ooinet.oceanobservatories.org/data_access/
?search=RS03AXBS-MJ03A-05-HYDLFA301>`_
    * 'Axial_Base'
    * 'AXABA1'
* `Central Caldera (Fs = 200 Hz)
<https://ooinet.oceanobservatories.org/data_access/
?search=RS03CCAL-MJ03F-06-HYDLFA305>`_
    * 'Central_Caldera'
    * 'AXCC1'
* `Eastern Caldera (Fs = 200 Hz)
<https://ooinet.oceanobservatories.org/data_access/
?search=RS03ECAL-MJ03E-09-HYDLFA304>`_
    * 'Eastern_Caldera'
    * 'AXEC2'
* `Southern Hydrate (Fs = 200 Hz)
<https://ooinet.oceanobservatories.org/data_access/
?search=RS01SUM1-LJ01B-05-HYDLFA104>`_
    * 'Southern_Hydrate'
    * 'HYS14'
* `Oregon Slope Base Seafloor (Fs = 200 Hz)
<https://ooinet.oceanobservatories.org/data_access/
?search=RS01SLBS-MJ01A-05-HYDLFA101>`_
    * 'Slope_Base'
    * 'HYSB1'

Hydrophone Request Modules
^^^^^^^^^^^^^^^^^^^^^^^^^^
'''

# Import all dependancies
from ooipy.hydrophone.basic import HydrophoneData
from obspy import read, Stream
from obspy.core import UTCDateTime
from datetime import timedelta
import numpy as np
import multiprocessing as mp
import concurrent.futures
import fsspec


def get_acoustic_data(starttime, endtime, node, fmin=None, fmax=None,
                      max_workers=-1, append=True, verbose=False,
                      data_gap_mode=0, mseed_file_limit=None,
                      large_gap_limit=1800.0):
    '''
    Get broadband acoustic data for specific time frame and sensor node. The
    data is returned as a :class:`.HydrophoneData` object. This object is
    based on the obspy data trace. Example usage is shown below. For a more in
    depth tutorial, see the `Hydrophone Request Jupyter Notebook Example
    <_static/test_request.html>`_.

    >>> import ooipy
    >>> start_time = datetime.datetime(2017,3,10,0,0,0)
    >>> end_time = datetime.datetime(2017,3,10,0,5,0)
    >>> node = 'PC01A'
    >>> data = ooipy.request.get_acoustic_data(start_time, end_time, node)
    >>> # To access stats for retrieved data:
    >>> print(data.stats)
    >>> # To access numpy array of data:
    >>> print(data.data)

    Parameters
    ----------
    start_time : datetime.datetime
        time of the first noise sample
    end_time : datetime.datetime
        time of the last noise sample
    node : str
        hydrophone name or identifier
    fmin : float, optional
        lower cutoff frequency of hydrophone's bandpass filter. Default
        is None which results in no filtering.
    fmax : float, optional
        higher cutoff frequency of hydrophones bandpass filter. Default
        is None which results in no filtering.
    print_exceptions : bool, optional
        whether or not exeptions are printed in the terminal line
    max_workers : int, optional
        number of maximum workers for concurrent processing
    append : bool, optional
        specifies if extra mseed files should be appended at beginning
        and end in case of boundary gaps in data
    verbose : bool, optional
        specifies whether print statements should occur or not
    data_gap_mode : int, optional
        How gaps in the raw data will be handled. Options are:
        '0': gaps will be linearly interpolated
        '1': no interpolation; mask array is returned
        '2': subtract mean of data and fill gap with zeros; mask array
        is returned
    mseed_file_limit: int, optional
        If the number of mseed files to be merged exceed this value, the
        function returns None. For some days the mseed files contain
        only a few seconds or milli seconds of data and merging a huge
        amount of files can dramatically slow down the program. if None
        (default), the number of mseed files will not be limited.
    large_gap_limit: float, optional
        Defines the length in second of large gaps in the data.
        Sometimes, large data gaps are present on particular days. This
        can cause long interpolation times if data_gap_mode 0 or 2 are
        used, possibly resulting in a memory overflow. If a data gap is
        longer than large_gap_limit, data are only retrieved before (if
        the gap stretches beyond the requested time) or after (if the gap
        starts prior to the requested time) the gap, or not at all (if
        the gap is within the requested time).

    Returns
    -------
    HydrophoneData

    '''

    # data_gap = False
    sampling_rate = 64000.0

    if verbose:
        print('Fetching URLs...')

    # get URL for first day
    day_start = UTCDateTime(starttime.year, starttime.month, starttime.day,
                            0, 0, 0)
    data_url_list = __get_mseed_urls(starttime.strftime("/%Y/%m/%d/"), node,
                                     verbose)

    if data_url_list is None:
        if verbose:
            print('No data available for specified day and node. '
                  'Please change the day or use a different node')
        return None

    # increment day start by 1 day
    day_start = day_start + 24 * 3600

    # get all urls for each day until endtime is reached
    while day_start < endtime:
        urls_list_next_day = __get_mseed_urls(day_start.strftime("/%Y/%m/%d/"),
                                              node, verbose)
        if urls_list_next_day is None:
            day_start = day_start + 24 * 3600
        else:
            data_url_list.extend(urls_list_next_day)
            day_start = day_start + 24 * 3600

    if append:
        # Save last mseed of previous day to data_url_list if not None
        prev_day = starttime - timedelta(days=1)
        data_url_list_prev_day = \
            __get_mseed_urls(prev_day.strftime("/%Y/%m/%d/"), node, verbose)
        if data_url_list_prev_day is not None:
            data_url_list = [data_url_list_prev_day[-1]] + data_url_list

        # get 1 more day of urls
        data_url_last_day_list = \
            __get_mseed_urls(day_start.strftime("/%Y/%m/%d/"), node, verbose)
        if data_url_last_day_list is not None:
            data_url_list = data_url_list + [data_url_last_day_list[0]]

    if verbose:
        print('Sorting valid URLs for Time Window...')
    # Create list of urls for specified time range

    valid_data_url_list = []
    first_file = True

    # Create List of mseed urls for valid time range
    for i in range(len(data_url_list)):

        # get UTC time of current and next item in URL list
        # extract start time from ith file
        utc_time_url_start = UTCDateTime(
            data_url_list[i].split('YDH')[1][1:].split('.mseed')[0])

        # this line assumes no gaps between current and next file
        if i != len(data_url_list) - 1:
            utc_time_url_stop = UTCDateTime(
                data_url_list[i + 1].split('YDH')[1][1:].split('.mseed')[0])
        else:
            utc_time_url_stop = UTCDateTime(
                data_url_list[i].split('YDH')[1][1:].split('.mseed')[0])
            utc_time_url_stop.hour = 23
            utc_time_url_stop.minute = 59
            utc_time_url_stop.second = 59
            utc_time_url_stop.microsecond = 999999

        # if current segment contains desired data, store data segment
        if (utc_time_url_start >= starttime
            and utc_time_url_start < endtime) \
            or (utc_time_url_stop >= starttime
                and utc_time_url_stop < endtime) \
            or (utc_time_url_start <= starttime
                and utc_time_url_stop >= endtime):
            if append:
                if i == 0:
                    first_file = False
                    valid_data_url_list.append(data_url_list[i])

                elif (first_file):
                    first_file = False
                    valid_data_url_list = [data_url_list[i - 1],
                                           data_url_list[i]]
                else:
                    valid_data_url_list.append(data_url_list[i])
            else:
                if i == 0:
                    first_file = False
                valid_data_url_list.append(data_url_list[i])

        # adds one more mseed file to st_ll
        else:
            # Checks if last file has been downloaded within time period
            if first_file is False:
                first_file = True
                if append:
                    valid_data_url_list.append(data_url_list[i])
                break

    if isinstance(mseed_file_limit, int):
        if len(valid_data_url_list) > mseed_file_limit:
            if verbose:
                print('Number of mseed files to be merged exceed limit.')
            return None

    # handle large data gaps within one day
    if len(valid_data_url_list) >= 2:
        # find gaps
        gaps = []
        for i in range(len(valid_data_url_list) - 1):
            utc_time_url_first = UTCDateTime(
                valid_data_url_list[i].split('YDH')[1][1:].split('.mseed')[0])
            utc_time_url_second = UTCDateTime(
                valid_data_url_list[i + 1].split('YDH')[1][1:].split(
                    '.mseed')[0])
            if utc_time_url_second - utc_time_url_first >= large_gap_limit:
                gaps.append(i)

        gap_cnt = 0
        # check if gap at beginning
        if 0 in gaps:
            del valid_data_url_list[0]
            gap_cnt += 1
            if verbose:
                print('Removed large data gap at beginning of requested time')
        # check if gap at the end
        if len(valid_data_url_list) - 2 in gaps:
            del valid_data_url_list[-1]
            gap_cnt += 1
            if verbose:
                print('Removed large data gap at end of requested time')
        # check if gap within requested time
        if len(gaps) > gap_cnt:
            if verbose:
                print('Found large data gap within requested time')
            return None

    if verbose:
        print('Downloading mseed files...')

    # Code Below from Landung Setiawan
    # removed max workers argument in following statement
    st_list = __map_concurrency(__read_mseed, valid_data_url_list)
    st_all = None
    for st in st_list:
        if st:
            if st[0].stats.sampling_rate != sampling_rate:
                if verbose:
                    print('Some data have different sampling rate')
            else:
                if not isinstance(st_all, Stream):
                    st_all = st
                else:
                    st_all += st

    if st_all is None:
        if verbose:
            print('No data available for selected time')
        return None

    # Merge all traces together
    # Interpolation
    if data_gap_mode == 0:
        st_all.merge(fill_value='interpolate', method=1)
    # Masked Array
    elif data_gap_mode == 1:
        st_all.merge(method=1)
    # Masked Array, Zero-Mean, Zero Fill
    elif data_gap_mode == 2:
        st_all.merge(method=1)
        st_all[0].data = st_all[0].data - np.mean(st_all[0].data)

        try:
            st_all[0].data.fill_value = 0
            st_all[0].data = np.ma.filled(st_all[0].data)
        except Exception:
            if verbose:
                print('data has no minor gaps')

    else:
        if verbose:
            print('Invalid Data Gap Mode')
        return None
    # Slice data to desired window
    st_all = st_all.slice(UTCDateTime(starttime), UTCDateTime(endtime))

    if len(st_all) == 0:
        if verbose:
            print('No data available for selected time frame.')
        return None

    if isinstance(st_all[0].data, np.ma.core.MaskedArray):
        # data_gap = True
        if verbose:  # Note this will only trip if masked array is returned
            print('Data has Gaps')
        # interpolated is treated as if there is no gap

    # Filter Data
    try:
        if fmin is not None and fmax is not None:
            st_all = st_all.filter("bandpass", freqmin=fmin,
                                   freqmax=fmax)
            if verbose:
                print('Signal Filtered')
        # return st_all[0]
        return HydrophoneData(st_all[0].data, st_all[0].stats, node)
    except Exception:
        if st_all is None:
            if verbose:
                print('No data available for selected time frame.')
        else:
            if verbose:
                print(Exception)
        return None


def get_acoustic_data_LF(starttime, endtime, node, fmin=None, fmax=None,
                         verbose=False, zero_mean=False):
    '''
    Get low frequency acoustic data for specific time frame and sensor
    node. The data is returned as a :class:`.HydrophoneData` object.
    This object is based on the obspy data trace. Example usage is shown
    below. For a more in depth tutorial, see the `Hydrophone Request
    Jupyter Notebook Example <_static/test_request.html>`_. This
    function does not include the full functionality provided by the
    `IRIS data portal
    <https://service.iris.edu/irisws/timeseries/docs/1/builder/>`_.


    >>> starttime = datetime.datetime(2017,3,10,7,0,0)
    >>> endtime = datetime.datetime(2017,3,10,7,1,30)
    >>> location = 'Axial_Base'
    >>> fmin = None
    >>> fmax = None
    >>> # Returns ooipy.ooipy.hydrophone.base.HydrophoneData Object
    >>> data_trace = hydrophone_request.get_acoustic_data_LF(
            starttime, endtime, location, fmin, fmax, zero_mean=True)
    >>> # Access data stats
    >>> data_trace.stats
    >>> # Access numpy array containing data
    >>> data_trace.data

    Parameters
    ----------
    start_time : datetime.datetime
        time of the first noise sample
    end_time : datetime.datetime
        time of the last noise sample
    node : str
        hydrophone
    fmin : float, optional
        lower cutoff frequency of hydrophone's bandpass filter. Default
        is None which results in no filtering.
    fmax : float, optional
        higher cutoff frequency of hydrophones bandpass filter. Default
        is None which results in no filtering.
    verbose : bool, optional
        specifies whether print statements should occur or not
    zero_mean : bool, optional
        specifies whether the mean should be removed. Default to False
    '''

    if fmin is None and fmax is None:
        bandpass_range = None
    else:
        bandpass_range = [fmin, fmax]

    url = __build_LF_URL(node, starttime, endtime,
                         bandpass_range=bandpass_range, zero_mean=zero_mean)
    if verbose:
        print('Downloading mseed file...')

    # Try downloading data 5 times. If fails every time raise exception
    for k in range(5):
        try:
            data_stream = read(url)
            break
        except Exception:
            if k == 4:
                print('   Specific Time window timed out.')
                return None

            # raise Exception ('Problem Requesting Data from OOI Server')

    hydrophone_data = HydrophoneData(data_stream[0].data, data_stream[0].stats,
                                     node)
    return hydrophone_data


def ooipy_read(device, node, starttime, endtime, fmin=None, fmax=None,
               verbose=False, data_gap_mode=0, zero_mean=False):
    '''
    General Purpose OOIpy read function. Parses input parameters to
    appropriate, device specific, read function. This function is under
    development but is included as is. There is no gurentee that this
    function works as expected.

    Parameters
    ----------
    device : str
        Specifies device type. Valid option are 'broadband_hydrohpone'
        and 'low_frequency_hydrophone'
    node : str
        Specifies data acquisition device location. TODO add available
        options
    starttime : datetime.datetime
        Specifies start time of data requested
    endtime : datetime.datetime
        Specifies end time of data requested
    fmin : float
        Low frequency corner for filtering. If None are give, then no
        filtering happens. Broadband hydrophone data is filtered using
        Obspy. Low frequency hydrophone uses IRIS filtering.
    fmax : float
        High frequency corner for filtering.
    verbose : bool
        Specifies whether or not to print status update statements.
    data_gap_mode : int
        specifies how gaps in data are handled see documentation for
        get_acoustic_data

    Returns
    -------
    hydrophone_data : HydrophoneData
        Object that stores hydrophone data. Similar to obspy trace.
    '''

    if device == 'broadband_hydrophone':
        hydrophone_data = get_acoustic_data(starttime, endtime, node, fmin,
                                            fmax, verbose=verbose,
                                            data_gap_mode=data_gap_mode)
    elif device == 'low_frequency_hydrophone':
        hydrophone_data = get_acoustic_data_LF(starttime, endtime, node,
                                               fmin=fmin, fmax=fmax,
                                               verbose=verbose,
                                               zero_mean=zero_mean)
    else:
        raise Exception('Invalid Devic String')

    return hydrophone_data


def __map_concurrency(func, iterator, args=(), max_workers=-1):
    # automatically set max_workers to 2x(available cores)
    if max_workers == -1:
        max_workers = 2 * mp.cpu_count()

    results = []
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) \
            as executor:
        # Start the load operations and mark each future with its URL
        future_to_url = {executor.submit(func, i, *args): i for i in iterator}
        for future in concurrent.futures.as_completed(future_to_url):
            data = future.result()
            results.append(data)
    return results


def __read_mseed(url):
    # fname = os.path.basename(url)
    try:
        st = read(url, apply_calib=True)
    except Exception:
        print(f'Data Segment {url} Broken')
        return None
    if isinstance(st, Stream):

        return st
    else:
        print(f'Problem Reading {url}')

        return None


def __get_mseed_urls(day_str, node, verbose):
    '''
    get URLs for a specific day from OOI raw data server

    Parameters
    ----------
    day_str : str
        date for which URLs are requested; format: yyyy/mm/dd,
        e.g. 2016/07/15
    node : str
        identifier or name of the hydrophone node
    verbose : bool
        print exceptions if True

    Returns
    -------
    ([str], str)
        list of URLs, each URL refers to one data file. If no data is
        available for specified date, None is returned.
    '''

    try:
        if node == 'LJ01D' or node == 'Oregon_Shelf_Base_Seafloor':
            array = '/CE02SHBP'
            instrument = '/11-HYDBBA106'
            node_id = '/LJ01D'
        if node == 'LJ01A' or node == 'Oregon_Slope_Base_Seafloor':
            array = '/RS01SLBS'
            instrument = '/09-HYDBBA102'
            node_id = '/LJ01A'
        if node == 'PC01A' or node == 'Oregon_Slope_Base_Shallow':
            array = '/RS01SBPS'
            instrument = '/08-HYDBBA103'
            node_id = '/PC01A'
        if node == 'PC03A' or node == 'Axial_Base_Shallow':
            array = '/RS03AXPS'
            instrument = '/08-HYDBBA303'
            node_id = '/PC03A'
        if node == 'LJ01C' or node == 'Oregon_Offshore_Base_Seafloor':
            array = '/CE04OSBP'
            instrument = '/11-HYDBBA105'
            node_id = '/LJ01C'
        if node == 'LJ03A' or node == 'Axial_Base_Seafloor':
            array = '/RS03AXBS'
            instrument = '/09-HYDBBA302'
            node_id = '/LJ03A'

        mainurl = 'https://rawdata.oceanobservatories.org/files' + array \
            + node_id + instrument + day_str
    except Exception:
        raise Exception('Invalid Location String ' + node + '. Please use one '
                        + 'of the following node strings: ' +
                        "'Oregon_Shelf_Base_Seafloor' ('LJ01D'); ",
                        "'Oregon_Slope_Base_Seafloor' ('LJ01A'); ",
                        "'Oregon_Slope_Base_Shallow' ('PC01A'); ",
                        "'Axial_Base_Shallow' ('PC03A'); ",
                        "'Oregon_Offshore_Base_Seafloor' ('LJ01C'); ",
                        "'Axial_Base_Seafloor' ('LJ03A')")

    FS = fsspec.filesystem('http')

    try:
        data_url_list = sorted([f['name'] for f in FS.ls(mainurl)
                               if f['type'] == 'file'
                               and f['name'].endswith('.mseed')])
    except Exception as e:
        if verbose:
            print('Client response: ', e)
        return None

    if not data_url_list:
        if verbose:
            print('No Data Available for Specified Time')
        return None

    return data_url_list


def __build_LF_URL(node, starttime, endtime, bandpass_range=None,
                   zero_mean=False, correct=False):
    '''
    Build URL for Lowfrequency Data given the start time, end time, and
    node

    Parameters
    ----------
    node : str
        node of low frequency hydrophone. Options include
        'Easter_Caldera', ...
    starttime : datetime.datetime
        start of data segment requested
    endtime : datetime.datetime
        end of data segment requested
    bandpass_range : list
        list of length two specifying [flow, fhigh] in Hertz. If None
        are given, no bandpass will be added to data.
    zero_mean : bool
        specified whether mean should be removed from data
    correct : bool
        specifies whether to do sensitivity correction on hydrophone data

    Returns
    -------
    url : str
        url of specified data segment. Format will be in miniseed.
    '''

    network, station, location, channel = __get_LF_locations_stats(node)

    starttime = starttime.strftime("%Y-%m-%dT%H:%M:%S")
    endtime = endtime.strftime("%Y-%m-%dT%H:%M:%S")
    base_url = 'http://service.iris.edu/irisws/timeseries/1/query?'
    netw_url = 'net=' + network + '&'
    stat_url = 'sta=' + station + '&'
    chan_url = 'cha=' + channel + '&'
    strt_url = 'start=' + starttime + '&'
    end_url = 'end=' + endtime + '&'
    form_url = 'format=miniseed&'
    loca_url = 'loc=' + location
    if correct:
        corr_url = '&correct=true'
    else:
        corr_url = ''

    if bandpass_range is None:
        band_url = ''
    else:
        band_url = 'bp=' + str(bandpass_range[0]) + '-' \
            + str(bandpass_range[1]) + '&'
    if zero_mean:
        mean_url = 'demean=true&'
    else:
        mean_url = ''
    url = base_url + netw_url + stat_url + chan_url + strt_url + end_url \
        + mean_url + band_url + form_url + loca_url + corr_url
    return url


def __get_LF_locations_stats(node):
    try:
        if node == 'Slope_Base' or node == 'HYSB1':
            network = 'OO'
            station = 'HYSB1'
            location = '--'
            channel = 'HDH'

        if node == 'Southern_Hydrate' or node == 'HYS14':
            network = 'OO'
            station = 'HYS14'
            location = '--'
            channel = 'HDH'

        if node == 'Axial_Base' or node == 'AXBA1':
            network = 'OO'
            station = 'AXBA1'
            location = '--'
            channel = 'HDH'

        if node == 'Central_Caldera' or node == 'AXCC1':
            network = 'OO'
            station = 'AXCC1'
            location = '--'
            channel = 'HDH'

        if node == 'Eastern_Caldera' or node == 'AXEC2':
            network = 'OO'
            station = 'AXEC2'
            location = '--'
            channel = 'HDH'

        # Create error if node is invalid
        network = network

    except Exception:
        raise Exception('Invalid Location String ' + node + '. Please use one '
                        + 'of the following node strings: ' +
                        "'Slope_Base' ('HYSB1'); ",
                        "'Southern_Hydrate' ('HYS14'); ",
                        "'Axial_Base' ('AXBA1'); ",
                        "'Central_Caldera' ('AXCC1'); ",
                        "'Eastern_Caldera' ('AXEC2')")

    return network, station, location, channel



import os.path
"""
This module provides functions for the automatic authentification at the
OOI. The authentification is based on a username and a token, which will
be generatd after registering an account at
https://ooinet.oceanobservatories.org/.
"""


def set_authentification(username, token):
    """
    Writes username and token to a text file. When requesting data from
    the OOI webservers, this textfile will automatically be accessed

    Parameters
    ----------
    username : str
        Username automatically generated by the OOI. It typically starts
        with "OOIAPI-...".
    token : str
        Token automatically generated by the OOI. It typically starts
        with "TEMP-TOKEN-...".
    """

    filename = 'ooi_auth.txt'
    if not os.path.isfile(filename):
        file = open(filename, 'w+')
        file.write('username\n' + username + '\n' + 'token\n' + token)
        file.close()


def get_authentification():
    """
    Open ooi_auth.txt file and return the username and the token

    Returns
    -------
    (str, str)
    """

    filename = 'ooi_auth.txt'
    if os.path.isfile(filename):
        file = open(filename, 'r')
        auth = file.readlines()
        username = auth[1].split('\n')[0]
        token = auth[3].split('\n')[0]
        file.close()
        return username, token



from ooipy.hydrophone.basic import Psd, Spectrogram
from ooipy.request.hydrophone_request import get_acoustic_data 
from ooipy.request.hydrophone_request import get_acoustic_data_LF
from ooipy.request.ctd_request import get_ctd_data 
from ooipy.request.ctd_request import get_ctd_data_daily
from ooipy.request.authentification import set_authentification
from ooipy.tools.ooiplotlib import plot
import ooipy.tools.workflow

from _ooipy_version import version as __version__



import requests
import multiprocessing as mp
import concurrent.futures
import datetime
import ooipy.request.authentification
from ooipy.ctd.basic import CtdData


def get_ctd_data(start_datetime, end_datetime, location, limit=10000,
                 only_profilers=False, delivery_method='auto'):
    """
    Requests CTD data between start_detetime and end_datetime for the
    specified location if data is available. For each location, data of
    all available CTDs are requested and concatenated in a list. That is
    the final data list can consists of multiple segments of data, where
    each segment contains the data from one instrument. This means that
    the final list might not be ordered in time, but rather should be
    treated as an unordered list of CTD data points.

    Parameters
    ----------
    start_datetime : datetime.datetime
        time of first sample from CTD
    end_datetime : datetime.datetime
        time of last sample from CTD
    location : str
        location for which data are requested. Possible choices are:
        * 'oregon_inshore'
        * 'oregon_shelf'
        * 'oregon_offshore'
        * 'oregon_slope'
        * 'washington_inshore'
        * 'washington_shelf'
        * 'washington_offshore'
        * 'axial_base'
    limit : int
        maximum number of data points returned in one request. The limit
        applies for each instrument separately. That is the final list
        of data points can contain more samples then indicated by limit
        if data from multiple CTDs is available at the given location
        and time. Default is 10,0000.
    only_profilers : bool
        Specifies whether only data from the water column profilers
        should be requested. Default is False
    delivery_method : str
        Specifies which delivery method is considered. For details
        please refer to http://oceanobservatories.org/glossary/. Options
        are:
        * 'auto' (default): automatically uses method that has data
            avaiilable
        * 'streamed': only consideres data that are streamed to shore
            via cable
        * 'telemetered': only consideres data that are streamed to shore
            via satellite
        * 'recovered': only consideres data that were reteived when the
            instrument was retreived

    Returns
    -------
    ctd_data : :class:`ooipy.ctd.basic.CtdProfile`
        object, where the data array is stored in the raw_data attribute. Each
        data sample consists of a dictionary of parameters measured by the CTD.

    """

    USERNAME, TOKEN = ooipy.request.authentification.get_authentification()
    # Sensor Inventory
    DATA_API_BASE_URL = \
        'https://ooinet.oceanobservatories.org/api/m2m/12576/sensor/inv/'

    # Oregon Shelf
    if location == 'oregon_shelf':
        url_list = \
            ['CE02SHSP/SP001/08-CTDPFJ000/telemetered/' +
             'ctdpf_j_cspp_instrument?',
             'CE02SHSP/SP001/08-CTDPFJ000/recovered_cspp/' +
             'ctdpf_j_cspp_instrument_recovered?']
        if not only_profilers:
            url_list.extend(['CE02SHBP/LJ01D/06-CTDBPN106/streamed/' +
                             'ctdbp_no_sample?',
                             'CE02SHSM/RID27/03-CTDBPC000/telemetered/' +
                             'ctdbp_cdef_dcl_instrument?'])

    elif location == 'oregon_offshore':
        url_list = \
            ['CE04OSPS/SF01B/2A-CTDPFA107/streamed/ctdpf_sbe43_sample?',
             'CE04OSPD/DP01B/01-CTDPFL105/recovered_inst/' +
             'dpc_ctd_instrument_recovered?',
             'CE04OSPD/DP01B/01-CTDPFL105/recovered_wfp/' +
             'dpc_ctd_instrument_recovered?']
        if not only_profilers:
            url_list.extend(['CE04OSPS/PC01B/4A-CTDPFA109/streamed/' +
                             'ctdpf_optode_sample?',
                             'CE04OSSM/RID27/03-CTDBPC000/telemetered/' +
                             'ctdbp_cdef_dcl_instrument?',
                             'CE04OSBP/LJ01C/06-CTDBPO108/streamed/' +
                             'ctdbp_no_sample?'])

    elif location == 'oregon_slope':
        url_list = \
            ['RS01SBPD/DP01A/01-CTDPFL104/recovered_inst/' +
             'dpc_ctd_instrument_recovered?',
             'RS01SBPD/DP01A/01-CTDPFL104/recovered_wfp/' +
             'dpc_ctd_instrument_recovered?',
             'RS01SBPS/SF01A/2A-CTDPFA102/streamed/ctdpf_sbe43_sample?']
        if not only_profilers:
            url_list.extend(['RS01SBPS/PC01A/4A-CTDPFA103/streamed/' +
                             'ctdpf_optode_sample?',
                             'RS01SLBS/LJ01A/12-CTDPFB101/streamed/' +
                             'ctdpf_optode_sample?'])

    elif location == 'oregon_inshore':
        url_list = \
            ['CE01ISSP/SP001/09-CTDPFJ000/recovered_cspp/' +
             'ctdpf_j_cspp_instrument_recovered?',
             'CE01ISSP/SP001/09-CTDPFJ000/telemetered/' +
             'ctdpf_j_cspp_instrument?']
        if not only_profilers:
            url_list.extend(['CE01ISSM/SBD17/06-CTDBPC000/telemetered/' +
                             'ctdbp_cdef_dcl_instrument?',
                             'CE01ISSM/SBD17/06-CTDBPC000/recovered_inst/' +
                             'ctdbp_cdef_instrument_recovered?',
                             'CE01ISSM/RID16/03-CTDBPC000/recovered_inst/' +
                             'ctdbp_cdef_instrument_recovered?',
                             'CE01ISSM/RID16/03-CTDBPC000/telemetered/' +
                             'ctdbp_cdef_dcl_instrument?',
                             'CE01ISSM/MFD37/03-CTDBPC000/recovered_inst/' +
                             'ctdbp_cdef_instrument_recovered?',
                             'CE01ISSM/MFD37/03-CTDBPC000/telemetered/' +
                             'ctdbp_cdef_dcl_instrument?'])

    elif location == 'washington_inshore':
        url_list = \
            ['CE06ISSP/SP001/09-CTDPFJ000/telemetered/' +
             'ctdpf_j_cspp_instrument?',
             'CE06ISSP/SP001/09-CTDPFJ000/recovered_cspp/' +
             'ctdpf_j_cspp_instrument_recovered?']
        if not only_profilers:
            url_list.extend(['CE06ISSM/SBD17/06-CTDBPC000/recovered_host/' +
                             'ctdbp_cdef_dcl_instrument_recovered?',
                             'CE06ISSM/SBD17/06-CTDBPC000/recovered_inst/' +
                             'ctdbp_cdef_instrument_recovered?',
                             'CE06ISSM/SBD17/06-CTDBPC000/telemetered/' +
                             'ctdbp_cdef_dcl_instrument?',
                             'CE06ISSM/RID16/03-CTDBPC000/recovered_inst/' +
                             'ctdbp_cdef_instrument_recovered?',
                             'CE06ISSM/RID16/03-CTDBPC000/telemetered/' +
                             'ctdbp_cdef_dcl_instrument?',
                             'CE06ISSM/MFD37/03-CTDBPC000/telemetered/' +
                             'ctdbp_cdef_dcl_instrument?',
                             'CE06ISSM/MFD37/03-CTDBPC000/recovered_inst/' +
                             'ctdbp_cdef_instrument_recovered?'])

    elif location == 'washington_shelf':
        url_list = \
            ['CE07SHSP/SP001/08-CTDPFJ000/recovered_cspp/' +
             'ctdpf_j_cspp_instrument_recovered?']
        if not only_profilers:
            url_list.extend(['CE07SHSM/RID27/03-CTDBPC000/telemetered/' +
                             'ctdbp_cdef_dcl_instrument?',
                             'CE07SHSM/RID27/03-CTDBPC000/recovered_inst/' +
                             'ctdbp_cdef_instrument_recovered?',
                             'CE07SHSM/MFD37/03-CTDBPC000/telemetered/' +
                             'ctdbp_cdef_dcl_instrument?',
                             'CE07SHSM/MFD37/03-CTDBPC000/recovered_inst/' +
                             'ctdbp_cdef_instrument_recovered?'])

    elif location == 'washington_offshore':
        url_list = \
            ['CE09OSPM/WFP01/03-CTDPFK000/telemetered/' +
             'ctdpf_ckl_wfp_instrument?',
             'CE09OSPM/WFP01/03-CTDPFK000/recovered_wfp/' +
             'ctdpf_ckl_wfp_instrument_recovered?']
        if not only_profilers:
            url_list.extend(['CE09OSSM/RID27/03-CTDBPC000/telemetered/' +
                             'ctdbp_cdef_dcl_instrument?',
                             'CE09OSSM/RID27/03-CTDBPC000/recovered_inst/' +
                             'ctdbp_cdef_instrument_recovered?',
                             'CE09OSSM/MFD37/03-CTDBPE000/telemetered/' +
                             'ctdbp_cdef_dcl_instrument?',
                             'CE09OSSM/MFD37/03-CTDBPE000/recovered_inst/' +
                             'ctdbp_cdef_instrument_recovered?'])

    elif location == 'axial_base':
        url_list = \
            ['RS03AXPD/DP03A/01-CTDPFL304/recovered_inst/' +
             'dpc_ctd_instrument_recovered?',
             'RS03AXPD/DP03A/01-CTDPFL304/recovered_wfp/' +
             'dpc_ctd_instrument_recovered?',
             'RS03AXPS/SF03A/2A-CTDPFA302/streamed/ctdpf_sbe43_sample?']
        if not only_profilers:
            url_list.extend(['RS03AXPS/PC03A/4A-CTDPFA303/streamed/' +
                             'ctdpf_optode_sample?',
                             'RS03AXBS/LJ03A/12-CTDPFB301/streamed/' +
                             'ctdpf_optode_sample?'])

    else:
        raise Exception('Do not know given location.')

    # remove URLs if delivery method is specified
    url_list_method = []
    if (delivery_method == 'telemetered' or delivery_method == 'streamed'
            or delivery_method == 'recovered'):
        for url in url_list:
            if delivery_method in url:
                url_list_method.append(url)
    elif delivery_method == 'auto':
        url_list_method = url_list
    else:
        raise Exception("deliveriy method need to be either 'auto', " +
                        "'telemetered', 'streamed', or 'recovered', but not " +
                        str(delivery_method))

    beginDT = start_datetime.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + 'Z'
    endDT = end_datetime.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + 'Z'

    dataraw = []
    processed_ctds = []

    for url in url_list_method:
        ctd_id = __get_instrument_id(url)
        if ctd_id in processed_ctds:
            continue

        data_request_url = DATA_API_BASE_URL + url + \
            'beginDT=' + beginDT + '&endDT=' + endDT + '&limit=' + str(limit)

        r = requests.get(data_request_url, auth=(USERNAME, TOKEN))
        dataraw2 = r.json()
        if 'message' not in dataraw2:
            dataraw.extend(dataraw2)
            processed_ctds.append(ctd_id)
    ctd_data = CtdData(raw_data=dataraw)
    return ctd_data


def __get_instrument_id(url_str):
    ctd_id_arr = url_str.split('/')[:3]
    ctd_id = ctd_id_arr[0] + '/' + ctd_id_arr[1] + '/' + ctd_id_arr[2]

    return ctd_id


def get_ctd_data_daily(datetime_day, location, limit=10000,
                       only_profilers=False, delivery_method='auto'):
    """
    Requests CTD data for specified day and location. The day is split
    into 24 1-hour periods and for each 1-hour period
    :func:`ooipy.ctd_request.get_ctd_data is called. The data for all
    1-hour periods are then concatednated in stored in a
    :class:`ooipy.ctd.basic.CtdProfile` object.

    Parameters
    ----------
    datetime_day : datetime.datetime
        Day for which CTD data are requested
    location : str
        See :func:`ooipy.ctd_request.get_ctd_data
    limit : int
        See :func:`ooipy.ctd_request.get_ctd_data
    only_profilers : bool
        See :func:`ooipy.ctd_request.get_ctd_data
    delivery_method : str
        See :func:`ooipy.ctd_request.get_ctd_data

    Returns
    -------
    :class:`ooipy.ctd.basic.CtdProfile` object, where the data array is
    stored in the raw_data attribute. Each data sample consists of a
    dictionary of parameters measured by the CTD.
    """
    # get CTD data for one hour
    year = datetime_day.year
    month = datetime_day.month
    day = datetime_day.day

    start_end_list = []

    for hour in range(24):
        start = datetime.datetime(year, month, day, hour, 0, 0)
        end = datetime.datetime(year, month, day, hour, 59, 59, 999)
        start_end_list.append((start, end))

    raw_data_arr = __map_concurrency(__get_ctd_data_concurrent, start_end_list,
                                     {'location': location, 'limit': limit,
                                      'only_profilers': only_profilers,
                                      'delivery_method': delivery_method})

    raw_data_falttened = []
    for item in raw_data_arr:
        if item is None:
            continue
        else:
            raw_data_falttened.extend(item.raw_data)

    return CtdData(raw_data=raw_data_falttened)


def __map_concurrency(func, iterator, args=(), max_workers=-1):
    """
    Helper function to support multiprocessing for
    :func:`ooipy.ctd_request.get_ctd_data_daily
    """
    # automatically set max_workers to 2x(available cores)
    if max_workers == -1:
        max_workers = 2 * mp.cpu_count()

    results = []
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) \
            as executor:
        # Start the load operations and mark each future with its URL
        future_to_url = {executor.submit(func, i, **args): i for i in iterator}
        for future in concurrent.futures.as_completed(future_to_url):
            data = future.result()
            results.append(data)
    return results


def __get_ctd_data_concurrent(start_end_tuple, location, limit, only_profilers,
                              delivery_method):
    """
    Helper function to support multiprocessing for
    :func:`ooipy.ctd_request.get_ctd_data_daily
    """
    start = start_end_tuple[0]
    end = start_end_tuple[1]

    rawdata = get_ctd_data(start, end, location=location, limit=limit,
                           only_profilers=only_profilers,
                           delivery_method=delivery_method)
    return rawdata



from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import os
import setuptools

here = os.path.abspath(os.path.dirname(__file__))

# Dependencies.
with open("requirements.txt") as f:
    requirements = f.readlines()
install_requires = [t.strip() for t in requirements]

with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="ooipy",
    author="OOIPy",
    author_email="ooipython@gmail.com",
    description="A python toolbox for acquiring and analyzing Ocean Obvservatories Initiative (OOI) Data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ooipy/ooipy",
    packages=setuptools.find_packages(exclude=("tests")),
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=install_requires,
    py_modules=["_ooipy_version"],
    use_scm_version={
        "write_to": "_ooipy_version.py",
        "write_to_template": 'version = "{version}"\n',
        "local_scheme": "no-local-version"
    },
    setup_requires=["setuptools>=30.3.0", "wheel", "setuptools_scm"]
)


"""
This modules provides some useful functions for saving spectrograms and
power spectral density objects.
"""

# Import all dependancies
import numpy as np
from ooipy.hydrophone.basic import Spectrogram, Psd
import json
import pickle


def save(save_obj, filename, **kwargs):
    """
    Save :class:`ooipy.hydrophone.basic.Spectrogram` and
    :class:`ooipy.hydrophone.basic.Psd` objects. Spectrograms along
    with ancillary data will be saved in a pickle file as a dictionary.
    PSD object along with ancillary data will be saved in a json file as
    a dictionary.

    Parameters
    ----------
    save_obj :
        :class:`ooipy.hydrophone.basic.Psd` or
        :class:`ooipy.hydrophone.basic.Spectrogram` object to be saved
    filename : str
        name of file.
    kwargs : int, float, or list for Psd and arbitrary for spectrogram
        Ancillary data saved with the objects. Keyword and value
        will be saved as dictionary entries
    """

    if isinstance(save_obj, Spectrogram):
        save_spectrogram(save_obj, filename, **kwargs)
    elif isinstance(save_obj, Psd):
        save_psd(save_obj, filename, **kwargs)
    else:
        raise Exception('Fuction only supports spectrogram and PSD objects.')


def save_psd(psd_obj, filename, **kwargs):
    """
    Save a :class:`ooipy.hydrophone.basic.Psd` object in a json file.

    Parameters
    ----------
    spec_obj : :class:`ooipy.hydrophone.basic.Psd`
        Psd object to be saved
    filename : str
        name of file.
    kwargs : int, float, or list
        Ancillary data saved with the PSD object. Keyword and value
        will be saved as dictionary entries
    """

    if len(psd_obj.freq) != len(psd_obj.values):
        f = np.linspace(0, len(psd_obj.values) - 1, len(psd_obj.values))
    else:
        f = psd_obj.freq

    if not isinstance(psd_obj.values, list):
        values = psd_obj.values.tolist()

    if not isinstance(f, list):
        f = f.tolist()

    dct = {'psd': values,
           'values': f}

    # store ancillary data in dictionary
    for key, value in kwargs.items():
        if isinstance(value, int) or isinstance(value, float):
            dct[key] = value
        elif not isinstance(value, list):
            dct[key] = value.tolist()
        else:
            dct[key] = value

    with open(filename, 'w+') as outfile:
        json.dump(dct, outfile)


def save_spectrogram(spectrogram_obj, filename, **kwargs):
    """
    Save a :class:`ooipy.hydrophone.basic.Spectrogram` object in a
    pickle file.

    Parameters
    ----------
    spectrogram_obj : :class:`ooipy.hydrophone.basic.Spectrogram`
        Spectrogram object to be saved
    filename : str
        name of file.
    kwargs : any datatype
        Ancillary data saved with the Spectrogram object. Keyword and
        value will be saved as dictionary entries
    """

    dct = {'time': spectrogram_obj.time,
           'frequency': spectrogram_obj.freq,
           'values': spectrogram_obj.values}

    for key, value in kwargs.items():
        dct[key] = value

    with open(filename, 'wb') as outfile:
        pickle.dump(dct, outfile)
